import math

import plotly.graph_objs as go
from plotly.subplots import make_subplots

from value_dashboard.reports.shared_plot_utils import *


@timed
def conversion_rate_line_plot(data: Union[pl.DataFrame, pd.DataFrame],
                              config: dict, options_panel: bool = True) -> pd.DataFrame:
    xplot_y_bool = False
    xplot_col = config.get('color', None)
    facet_row = '---' if not 'facet_row' in config.keys() else config['facet_row']
    facet_column = '---' if not 'facet_column' in config.keys() else config['facet_column']
    x_axis = config.get('x', None)
    y_axis = config.get('y', None)
    height = config.get('height', 640)

    if options_panel:
        toggle1, toggle2 = st.columns(2)
        cards_on = toggle1.toggle("Metric totals", value=True, key="Metrics" + config['description'],
                                  help="Show aggregated metric values with difference from mean")
        adv_on = toggle2.toggle("Advanced options", value=False, key="Advanced options" + config['description'],
                                help="Show advanced reporting options")

        if adv_on:
            plot_menu = get_plot_parameters_menu(config=config, is_y_axis_required=False)
            x_axis = plot_menu['x']
            y_axis = plot_menu['y']
            facet_column = plot_menu['facet_col']
            facet_row = plot_menu['facet_row']
            xplot_col = plot_menu['color']
            xplot_y_bool = plot_menu['log_y']

    grp_by = [x_axis]
    if not (facet_column == '---'):
        if not facet_column in grp_by:
            grp_by.append(facet_column)
    else:
        facet_column = None
    if not (facet_row == '---'):
        if not facet_row in grp_by:
            grp_by.append(facet_row)
    else:
        facet_row = None

    if not xplot_col in grp_by:
        grp_by.append(xplot_col)

    cp_config = config.copy()
    cp_config['group_by'] = grp_by

    ih_analysis = data.copy()
    if options_panel:
        ih_analysis = filter_dataframe(align_column_types(ih_analysis), case=False)
    ih_analysis = calculate_reports_data(ih_analysis, cp_config).to_pandas()
    if ih_analysis.shape[0] == 0:
        st.warning("No data available.")
        return ih_analysis
    if options_panel and cards_on:
        conversion_rate_cards_subplot(ih_analysis, cp_config)

    if len(ih_analysis[x_axis].unique()) < 30:
        fig = px.bar(ih_analysis,
                     x=x_axis,
                     y=y_axis,
                     color=xplot_col,
                     error_y='StdErr',
                     facet_col=facet_column,
                     facet_row=facet_row,
                     barmode="group",
                     title=config['description'],
                     custom_data=[xplot_col],
                     log_y=xplot_y_bool
                     )
        fig.update_layout(
            updatemenus=[
                dict(
                    buttons=list([
                        dict(
                            args=["type", "bar"],
                            label="Bar",
                            method="restyle"
                        ),
                        dict(
                            args=["type", "line"],
                            label="Line",
                            method="restyle"
                        )
                    ]),
                    direction="down",
                    showactive=True,
                ),
            ]
        )
    else:
        fig = px.line(
            ih_analysis,
            x=x_axis,
            y=y_axis,
            color=xplot_col,
            title=config['description'],
            facet_col=facet_column,
            facet_row=facet_row,
            custom_data=[xplot_col],
            log_y=xplot_y_bool,
        )

    fig.update_xaxes(tickfont=dict(size=10))
    fig.update_yaxes(tickformat=',.2%')
    yaxis_names = ['yaxis'] + [axis_name for axis_name in fig.layout._subplotid_props if 'yaxis' in axis_name]
    yaxis_layout_dict = {yaxis_name + "_tickformat": ',.2%' for yaxis_name in yaxis_names}
    fig.update_layout(yaxis_layout_dict)
    if facet_row:
        height = max(640, 300 * len(ih_analysis[facet_row].unique()))
    fig.update_layout(
        xaxis_title=x_axis,
        yaxis_title=y_axis,
        hovermode="x unified",
        height=height
    )
    fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
    fig = fig.update_traces(hovertemplate=x_axis + ' : %{x}' + '<br>' +
                                          xplot_col + ' : %{customdata[0]}' + '<br>' +
                                          y_axis + ' : %{y:.2%}' + '<extra></extra>')
    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    return ih_analysis


@timed
def conversion_rate_gauge_plot(data: Union[pl.DataFrame, pd.DataFrame],
                               config: dict) -> pd.DatetimeIndex | None:
    report_data = calculate_reports_data(data, config).to_pandas()
    ih_analysis = filter_dataframe(align_column_types(report_data), case=False)
    if ih_analysis.shape[0] == 0:
        st.warning("No data available.")
        return ih_analysis
    grp_by = config['group_by']
    ih_analysis = ih_analysis.sort_values(by=grp_by)
    ih_analysis = ih_analysis.reset_index()

    if len(grp_by) == 0:
        st.warning("Group By property for Gauge plot should not be empty.")
        st.stop()
    elif len(grp_by) > 2:
        st.warning("Gauge plot type does not support more than two grouping columns.")
        st.stop()
    elif len(grp_by) == 2:
        cols = ih_analysis[grp_by[0]].unique().shape[0]
        rows = ih_analysis[grp_by[1]].unique().shape[0]
    else:
        cols = math.isqrt(ih_analysis[grp_by[0]].unique().shape[0])
        rows = cols + 1

    reference = config['reference']
    ih_analysis['Name'] = ih_analysis[grp_by].apply(lambda r: ' '.join(r.values.astype(str)), axis=1)
    ih_analysis['CName'] = ih_analysis[grp_by].apply(lambda r: '_'.join(r.values.astype(str)), axis=1)
    fig = make_subplots(rows=rows,
                        cols=cols,
                        specs=[[{"type": "indicator"} for c in range(cols)] for t in range(rows)]
                        )
    fig.update_layout(
        height=300 * rows,
        autosize=True,
        title=config['description'],
        margin=dict(b=10, t=120, l=10, r=10))
    for index, row in ih_analysis.iterrows():
        ref_value = reference.get(row['CName'], None)
        max_value = 1.1 * ih_analysis[ih_analysis[grp_by[0]] == row[grp_by[0]]][config['value']].max()
        gauge = {
            'axis': {'range': [None, max_value], 'tickformat': ',.2%'},
            'threshold': {
                'line': {'color': "red", 'width': 2},
                'thickness': 0.75,
                'value': ref_value
            }
        }
        if ref_value:
            if row[config['value']] < ref_value:
                gauge = {
                    'axis': {'range': [None, max_value], 'tickformat': ',.2%'},
                    'bar': {'color': '#EC5300' if row[config['value']] < (0.75 * ref_value) else '#EC9B00'},
                    'threshold': {
                        'line': {'color': "red", 'width': 2},
                        'thickness': 0.75,
                        'value': ref_value
                    }
                }

        trace1 = go.Indicator(mode="gauge+number+delta",
                              number={'valueformat': ",.2%"},
                              value=row[config['value']],
                              delta={'reference': ref_value, 'valueformat': ",.2%"},
                              title={'text': row['Name']},
                              gauge=gauge,
                              )
        r, c = divmod(index, cols)
        fig.add_trace(
            trace1,
            row=(r + 1), col=(c + 1)
        )
    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    ih_analysis.drop(columns=['Name', 'CName', 'index'], inplace=True, errors='ignore')
    return ih_analysis


@timed
def conversion_rate_cards_subplot(ih_analysis: Union[pl.DataFrame, pd.DataFrame],
                                  config: dict):
    grp_by = config['group_by']
    if len(grp_by) > 1:
        grp_by = grp_by[-2:]
    else:
        if len(grp_by) > 0:
            grp_by = grp_by[-1:]
    if isinstance(ih_analysis, pd.DataFrame):
        dfg = ih_analysis.groupby(grp_by)
        ih_analysis = pl.from_pandas(ih_analysis)
    else:
        dfg = ih_analysis.to_pandas().groupby(grp_by)
    if dfg.ngroups > 18:
        grp_by = config['group_by'][-1:]
    data_copy = (
        ih_analysis
        .group_by(grp_by)
        .agg(pl.sum("Negatives").alias("Negatives"),
             pl.sum("Positives").alias("Positives"))
        .with_columns([
            (pl.col("Positives") / (pl.col("Positives") + pl.col("Negatives"))).alias("ConversionRate")
        ])
        .sort(grp_by)
    )
    average = data_copy.select(((pl.col("Positives") + pl.col("Negatives")).dot(pl.col("ConversionRate"))) / (
        (pl.col("Positives") + pl.col("Negatives"))).sum()).item()
    data_copy = data_copy.to_pandas()
    num_metrics = data_copy.shape[0]
    dims = st.session_state['dashboard_dims']
    if dims:
        main_area_width = dims.get('width')
        max_num_cols = main_area_width // 120
    else:
        max_num_cols = 8
    num_cols = num_metrics if num_metrics < max_num_cols else max_num_cols
    cols = st.columns(num_cols)
    for index, row in data_copy.iterrows():
        if len(grp_by) > 1:
            kpi_name = row.iloc[0] + "  \n" + row.iloc[1]
        else:
            kpi_name = row.iloc[0]
        cols[index % max_num_cols].metric(label=kpi_name, value='{:.2%}'.format(row["ConversionRate"]),
                                          delta='{:.2%}'.format(row["ConversionRate"] - average))
        if (index + 1) % max_num_cols == 0:
            cols = st.columns(max_num_cols)


@timed
def conversion_rate_card(ih_analysis: Union[pl.DataFrame, pd.DataFrame]):
    data_copy = (
        ih_analysis
        .select(["Positives", "Negatives", "Revenue"])
        .sum()
        .with_columns([
            (pl.col("Positives") / (pl.col("Positives") + pl.col("Negatives"))).alias("ConversionRate")
        ])
    )
    st.metric(label="**Conversion**", value='{:.2%}'.format(data_copy["ConversionRate"].item()), border=False,
              help=f'Conversion rate', delta='Revenue {:,.0f}'.format(data_copy["Revenue"].item()), delta_color='off')


@timed
def conversion_touchpoints_card(ih_analysis: Union[pl.DataFrame, pd.DataFrame]):
    data_copy = (
        ih_analysis
        .select(["Positives", "Touchpoints"])
        .sum()
        .with_columns([
            (pl.col("Touchpoints") / pl.col("Positives")).alias("AvgTouchpoints")
        ])
    )
    st.metric(label="**Total conversions**", value='{:,.0f}'.format(data_copy["Positives"].item()), border=False,
              help=f'Conversions and touchpoints', delta=f'Avg Touchpoints: {data_copy["AvgTouchpoints"].item():.2}',
              delta_color='off')


@timed
def conversion_revenue_line_plot(data: Union[pl.DataFrame, pd.DataFrame],
                                 config: dict) -> pd.DataFrame:
    adv_on = st.toggle("Advanced options", value=False, key="Advanced options" + config['description'],
                       help="Show advanced reporting options")
    xplot_y_bool = False
    xplot_col = config.get('color', None)
    facet_row = '---' if not 'facet_row' in config.keys() else config['facet_row']
    facet_column = '---' if not 'facet_column' in config.keys() else config['facet_column']
    x_axis = config.get('x', None)
    y_axis = config.get('y', None)
    height = config.get('height', 640)

    if adv_on:
        plot_menu = get_plot_parameters_menu(config=config, is_y_axis_required=False)
        x_axis = plot_menu['x']
        y_axis = plot_menu['y']
        facet_column = plot_menu['facet_col']
        facet_row = plot_menu['facet_row']
        xplot_col = plot_menu['color']
        xplot_y_bool = plot_menu['log_y']

    grp_by = [x_axis]
    if not (facet_column == '---'):
        if not facet_column in grp_by:
            grp_by.append(facet_column)
    else:
        facet_column = None
    if not (facet_row == '---'):
        if not facet_row in grp_by:
            grp_by.append(facet_row)
    else:
        facet_row = None

    if not xplot_col in grp_by:
        grp_by.append(xplot_col)

    cp_config = config.copy()
    cp_config['group_by'] = grp_by

    ih_analysis = data.copy()
    ih_analysis = filter_dataframe(align_column_types(ih_analysis), case=False)
    ih_analysis = calculate_reports_data(ih_analysis, cp_config).to_pandas()
    if ih_analysis.shape[0] == 0:
        st.warning("No data available.")
        return ih_analysis

    if len(ih_analysis[x_axis].unique()) < 30:
        fig = px.bar(ih_analysis,
                     x=x_axis,
                     y=y_axis,
                     color=xplot_col,
                     facet_col=facet_column,
                     facet_row=facet_row,
                     barmode="group",
                     log_y=xplot_y_bool,
                     title=config['description']
                     )
        fig.update_xaxes(tickfont=dict(size=10))
        fig.update_yaxes(tickformat=',.2f')
        fig.update_layout(
            updatemenus=[
                dict(
                    buttons=list([
                        dict(
                            args=["type", "bar"],
                            label="Bar",
                            method="restyle"
                        ),
                        dict(
                            args=["type", "line"],
                            label="Line",
                            method="restyle"
                        )
                    ]),
                    direction="down",
                    showactive=True,
                ),
            ]
        )
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
    else:
        fig = px.line(
            ih_analysis,
            x=x_axis,
            y=y_axis,
            color=xplot_col,
            facet_col=facet_column,
            facet_row=facet_row,
            log_y=xplot_y_bool,
            title=config['description']
        )
        fig.update_xaxes(tickfont=dict(size=10))
        fig.update_yaxes(dict(tickformat=',.2'))
        fig.update_layout(
            xaxis_title=x_axis,
            yaxis_title=y_axis,
            hovermode="x unified",
            autosize=True,
            minreducedheight=height,
            height=height
        )
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    return ih_analysis



import os
import tempfile
import tomllib
import uuid
from traceback import print_stack

import polars as pl
import streamlit as st
import tomlkit
from pandasai.helpers.memory import Memory
from pandasai_openai import OpenAI

from value_dashboard.metrics.constants import DROP_IH_COLUMNS, OUTCOME_TIME, DECISION_TIME
from value_dashboard.utils.config import get_config, set_config
from value_dashboard.utils.file_utils import read_dataset_export
from value_dashboard.utils.logger import get_logger
from value_dashboard.utils.polars_utils import schema_with_unique_counts
from value_dashboard.utils.string_utils import capitalize

logger = get_logger(__name__)


@st.fragment()
def generate_new_config(llm, prompt):
    memory = Memory(agent_description="Config file generator.")
    new_config_text = llm.chat_completion(value=prompt, memory=memory)
    lines = new_config_text.splitlines(keepends=True)
    new_config_text = ''.join(lines[1:])
    new_config_text = new_config_text.replace('```', '')
    new_cfg = tomllib.loads(new_config_text)
    new_cfg["chat_with_data"] = get_config()["chat_with_data"]
    new_cfg["ux"]['chat_with_data'] = 'true'
    new_config_text = tomlkit.dumps(new_cfg)
    try:
        os.makedirs("temp_configs")
    except FileExistsError:
        pass
    cfg_file_name = "temp_configs/" + "config_" + uuid.uuid4().hex + '.toml'
    with open(cfg_file_name, "w") as f:
        f.write(new_config_text)

    set_config(cfg_file_name)
    st.download_button(
        label="Download",
        data=new_config_text,
        file_name="config.toml",
        mime="text/plain",
        type='primary'
    )


f"""## ✨ GenAI Config Generator"""
with st.sidebar:
    package_dir = os.path.dirname(__file__)
    template_config_file = os.path.join(package_dir, "../config", "config_template.toml")
    try:
        with open(template_config_file, mode="rb") as fp:
            template_config = tomllib.load(fp)
    except FileNotFoundError:
        print_stack()
        st.error(f"Configuration file not found: {template_config_file}")
        st.stop()
    except tomllib.TOMLDecodeError as e:
        print_stack()
        st.error(f"Configuration file is not valid TOML. {e}")
        st.stop()

    api_key_input = st.text_input(
        "Enter API Key (Leave empty to use environment variable)",
        type="password",
        value=os.environ.get("OPENAI_API_KEY"),
    )
    st.markdown(
        """
    <style>
        [title="Show password text"] {
            display: none;
        }
    </style>
    """,
        unsafe_allow_html=True,
    )
    openai_api_key = (
        api_key_input if api_key_input else os.environ.get("OPENAI_API_KEY")
    )
    if not openai_api_key:
        st.error("Please configure LLM API key.")
        st.stop()
    model_choice = st.selectbox(
        "Choose Model",
        options=OpenAI._supported_chat_models,
        index=OpenAI._supported_chat_models.index(OpenAI.model)
    )
    llm = OpenAI(
        api_token=openai_api_key,
        model=model_choice
    )

st.subheader("Choose file with IH sample", divider='red')
uploaded_file = st.file_uploader("*", type=["zip", "parquet", "json", "gzip"],
                                 accept_multiple_files=False)
df = pl.DataFrame()
if uploaded_file:
    temp_dir = tempfile.TemporaryDirectory(prefix='tmp')
    folder_path = os.path.abspath(temp_dir.name)
    file_path = os.path.join(temp_dir.name, uploaded_file.name)
    with open(file_path, "wb") as f:
        f.write(uploaded_file.getbuffer())
    df = read_dataset_export(file_names=uploaded_file.name, src_folder=os.path.dirname(file_path), lazy=False)
    dframe_columns = df.collect_schema().names()
    capitalized = capitalize(dframe_columns)
    rename_map = dict(zip(dframe_columns, capitalized))
    df = df.rename(rename_map)
    temp_dir.cleanup()

if not df.is_empty():
    df = df.lazy()
    with_cols_list = []
    if 'default_values' in template_config["ih"]["extensions"].keys():
        default_values = template_config["ih"]["extensions"]["default_values"]
        for new_col in default_values.keys():
            if new_col not in capitalized:
                with_cols_list.append(pl.lit(default_values.get(new_col)).alias(new_col))
            else:
                with_cols_list.append(pl.col(new_col).fill_null(default_values.get(new_col)))
    if with_cols_list:
        df = df.with_columns(with_cols_list)

    df = (
        df.with_columns([
            pl.col(OUTCOME_TIME).str.strptime(pl.Datetime, "%Y%m%dT%H%M%S%.3f %Z"),
            pl.col(DECISION_TIME).str.strptime(pl.Datetime, "%Y%m%dT%H%M%S%.3f %Z")
        ])
        .with_columns([
            pl.col(OUTCOME_TIME).dt.date().alias("Day"),
            pl.col(OUTCOME_TIME).dt.strftime("%Y-%m").alias("Month"),
            pl.col(OUTCOME_TIME).dt.year().cast(pl.Utf8).alias("Year"),
            (pl.col(OUTCOME_TIME).dt.year().cast(pl.Utf8) + "_Q" +
             pl.col(OUTCOME_TIME).dt.quarter().cast(pl.Utf8)).alias("Quarter"),
            (pl.col(OUTCOME_TIME) - pl.col(DECISION_TIME)).dt.total_seconds().alias("ResponseTime")
        ])
        .drop(DROP_IH_COLUMNS, strict=False)
        .collect()
    )
    df = df.select(sorted(df.columns))
    schema_df = schema_with_unique_counts(df).sort('Column')
    st.subheader("Schema", divider=True)
    st.data_editor(schema_df,
                   use_container_width=True,
                   disabled=True, height=300, hide_index=True)
    st.subheader("Data Summary", divider=True)
    st.dataframe(df.describe())

    with st.expander("View Data Sample", expanded=False, icon=":material/analytics:"):
        st.dataframe(df.head(100))

    with pl.Config(tbl_cols=len(schema_df), tbl_rows=len(schema_df)):
        prompt = f"""
            Given interaction history dataset schema (column names and types) and configuration file template, please create 
            similar config file, suited for this data. 
            Keep all the reports, metrics and other settings, but adjust columns, so they correspond to the 
            data in the file provided. Check columns available in the schema and include in the configuration 
            only those available in the sample. Do not generate 'chat_with_data' section.
            Replace names in template (in filters, group-by and report parameters) with column names in the schema (they may differ by case or have different prefixes or suffixes).
            Set 'file_type' to either 'parquet' (use file name extension to determine file type) or 'pega_ds_export' otherwise.
            Set 'file_pattern' extension accordingly.
            If column of type 'String' is not identifier column (ends with ID) or 'Outcome' and has number of unique values more than 1 and less than 100 - 
            append this column name to 'group_by' property of each metric in 'metrics' section.
            Always include time columns (Year, Quarter, Month, Day) and global filters to 'group_by' parameters.
            File name: {str(uploaded_file.name)}.
            Dataset schema: {schema_df}.
            Template config file: {tomlkit.dumps(template_config)}.
            """

        st.write("## Config from sample")
        if st.button("Generate config", key='UploadBtn', type='primary'):
            logger.debug('LLM prompt: ' + prompt)
            with st.spinner("Generating config. Wait for it...", show_time=True):
                logger.info('LLM call: ' + prompt)
                generate_new_config(llm, prompt)



import asyncio
import glob
import os
import re
import time
import typing
from collections import defaultdict
from datetime import timedelta

import numpy as np
import polars as pl
import streamlit as st
from polars import LazyFrame, DataFrame

from value_dashboard.pipeline.datatools import collect_clv_metrics_data
from value_dashboard.utils.config import get_config
from value_dashboard.utils.file_utils import read_dataset_export, detect_delimiter
from value_dashboard.utils.logger import get_logger
from value_dashboard.utils.string_utils import capitalize
from value_dashboard.utils.string_utils import strtobool
from value_dashboard.utils.timer import timed

HOLDINGS_FOLDER = "holdingsfolder"
logger = get_logger(__name__)
data_cache_hours = 24
if 'data_cache_hours' in get_config()['ux'].keys():
    data_cache_hours = get_config()['ux']['data_cache_hours']
logger.debug(f"Data will be cached for {data_cache_hours} hours.")
logger.debug(f"Numpy version {np.__version__}.")


def read_holdings_file_group(files: typing.Iterable,
                             filetype: str,
                             streaming: bool,
                             config: dict,
                             hive_partitioning: bool) -> LazyFrame | None:
    product_holdings_data_list = []
    start: float = time.time()
    for file in files:
        if filetype == 'parquet':
            product_holdings = pl.scan_parquet(file, cache=False, hive_partitioning=hive_partitioning,
                                               missing_columns='insert')
        elif filetype == 'pega_ds_export':
            product_holdings = read_dataset_export(file, lazy=True)
        elif filetype == 'csv':
            product_holdings = pl.scan_csv(
                source=file,
                separator=detect_delimiter(file),
                infer_schema_length=10000,
                try_parse_dates=True,
                cache=False
            )
        elif filetype == 'xlsx':
            product_holdings = pl.read_excel(source=file).lazy()
        else:
            raise Exception("File type not supported")

        logger.debug(f"Holdings data unpacking and load: {(time.time() - start) * 10 ** 3:.03f}ms")

        dframe_columns = product_holdings.collect_schema().names()
        cols = capitalize(dframe_columns)
        product_holdings = product_holdings.rename(dict(map(lambda i, j: (i, j), dframe_columns, cols)))
        product_holdings.collect_schema()

        if 'default_values' in config["holdings"]["extensions"].keys():
            default_values = config["holdings"]["extensions"]["default_values"]
            for new_col in default_values.keys():
                if new_col not in cols:
                    product_holdings = product_holdings.with_columns(pl.lit(default_values.get(new_col)).alias(new_col))
                else:
                    product_holdings = product_holdings.with_columns(
                        pl.col(new_col).fill_null(default_values.get(new_col)))

        global_ih_filter = config["holdings"]["extensions"]["filter"]
        if global_ih_filter:
            ih_filter_expr = eval(global_ih_filter)
            product_holdings = product_holdings.filter(ih_filter_expr)

        product_holdings_data_list.append(product_holdings)

    if not product_holdings_data_list:
        return
    product_holdings_group = pl.concat(product_holdings_data_list, how="diagonal")

    add_columns = config["holdings"]["extensions"]["columns"]
    if add_columns:
        add_columns_expr = eval(add_columns)
        product_holdings_group = product_holdings_group.with_columns(add_columns_expr)

    product_holdings_group = product_holdings_group.collect(engine="streaming" if streaming else "auto").lazy()
    logger.debug(f"Product holdings pre-processing took: {(time.time() - start) * 10 ** 3:.03f}ms")
    return product_holdings_group


@st.cache_data(show_spinner=False, ttl=timedelta(hours=data_cache_hours))
def load_holdings_data() -> typing.Dict[str, pl.DataFrame]:
    load_start = time.time()
    if HOLDINGS_FOLDER not in st.session_state:
        st.warning("Please configure your product holdings files in the `Data import` tab.")
        st.stop()
    folder = st.session_state[HOLDINGS_FOLDER]
    if not os.path.isdir(folder):
        st.error(f"Folder {folder} not available anymore. Please update data import parameters.")
        st.stop()
    file_groups = defaultdict(set)
    config = get_config()
    filetype = config["holdings"]["file_type"]
    logger.debug("File type: " + filetype)
    streaming = False
    if "streaming" in config['holdings'].keys():
        streaming = strtobool(config['holdings']['streaming'])
    logger.debug("Use polars streaming dataframe collect: " + str(streaming))
    background = False
    if "background" in config['holdings'].keys():
        background = strtobool(config['holdings']['background'])
    logger.debug("Use polars background dataframe collect: " + str(background))
    hive_partitioning = False
    if "hive_partitioning" in config['holdings'].keys():
        hive_partitioning = strtobool(config['holdings']['hive_partitioning'])
    logger.debug("Use hive partitioning: " + str(hive_partitioning))

    metrics = config["metrics"]
    for metric in metrics:
        params = metrics[metric]
        if isinstance(params, dict):
            if "filter" in params.keys():
                filter_exp_cmp = params["filter"]
                if isinstance(filter_exp_cmp, str):
                    if filter_exp_cmp:
                        params["filter"] = eval(filter_exp_cmp)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    files = [file for file in glob.iglob(folder + config["holdings"]["file_pattern"], recursive=True)]
    if not files:
        files = [file for file in glob.iglob(folder + '**/*.json', recursive=True)]
        filetype = 'pega_ds_export'

    for file in files:
        try:
            filedate = re.findall(config["holdings"]["file_group_pattern"], os.path.abspath(file))[0]
            if hive_partitioning:
                file_groups[filedate].add(os.path.dirname(file))
            else:
                file_groups[filedate].add(os.path.abspath(file))
        except Exception as e:
            file_groups[os.path.basename(file)].add(os.path.abspath(file))

    file_groups = dict(sorted(file_groups.items(), reverse=True))
    size = len(file_groups.items())
    mdata = {}
    progress_bar = st.progress(0)
    container = st.empty()
    i = 0
    for key, files in file_groups.items():
        logger.debug(f"Processing product holdings group: {key}")
        start = time.time()
        i = i + 1
        progress_bar.progress(i / size, text=f"Processing  product holdings: {key}")
        holdings_group = read_holdings_file_group(files, filetype, streaming, config, hive_partitioning)

        if holdings_group is None:
            continue

        collect_clv_metrics_data(loop, holdings_group, mdata, streaming, background, config)
        end = time.time()
        logger.debug(f"Product holdings time taken: {(end - start) * 10 ** 3:.03f}ms")
        load_mid = time.time()
        hours, remainder = divmod(load_mid - load_start, 3600)
        minutes, seconds = divmod(remainder, 60)
        container.metric("Time", f"{hours:.0f}h:{minutes:.0f}m:{seconds:.02f}s", label_visibility="collapsed")

    progress_bar.empty()
    container.empty()
    collected_metrics_data = {}

    for metric in mdata:
        totals_frame = mdata[metric]
        totals_frame.shrink_to_fit(in_place=True)
        collected_metrics_data[metric] = totals_frame

    load_end = time.time()
    hours, remainder = divmod(load_end - load_start, 3600)
    minutes, seconds = divmod(remainder, 60)
    logger.info(f"Load product holdings time: {hours:.0f}h:{minutes:.0f}m:{seconds:.02f}s")
    for metric in collected_metrics_data:
        frame = collected_metrics_data[metric]
        logger.debug(f"{metric} dataset size is {frame.estimated_size('kb')} kb.")
    return collected_metrics_data


@st.cache_data(show_spinner=False, ttl=timedelta(hours=data_cache_hours))
def get_reports_data():
    return collect_clv_reports_data(load_holdings_data())


@timed
def collect_clv_reports_data(collected_metrics_data: typing.Dict[str, pl.DataFrame]) -> dict[
    typing.Any, tuple[pl.DataFrame, typing.Any]]:
    report_params = get_config()["reports"]
    reports_data: dict[typing.Any, tuple[DataFrame, typing.Any]] = {}
    for report in report_params:
        params = report_params[report]
        if params['metric'].startswith("clv"):
            reports_data[report] = (collected_metrics_data[params['metric']], params)
    return reports_data



from value_dashboard.metrics.constants import PROPENSITY, FINAL_PROPENSITY
from value_dashboard.reports.repdata import calculate_model_ml_scores
from value_dashboard.reports.shared_plot_utils import *


@timed
def model_ml_scores_line_plot(data: Union[pl.DataFrame, pd.DataFrame],
                              config: dict) -> pd.DataFrame:
    ih_analysis = data.copy()
    ih_analysis = filter_dataframe(align_column_types(ih_analysis), case=False)
    ih_analysis = calculate_reports_data(ih_analysis, config).to_pandas()
    if ih_analysis.shape[0] == 0:
        st.warning("No data available.")
        return ih_analysis
    y_axis = config.get('y', None)
    x_axis = config.get('x', None)
    fig = px.line(
        ih_analysis,
        x=x_axis,
        y=y_axis,
        color=config['color'],
        log_y=config.get('log_y', False),
        title=config['description'],
        facet_row=config['facet_row'] if 'facet_row' in config.keys() else None,
        facet_col=config['facet_column'] if 'facet_column' in config.keys() else None,
        custom_data=[config['color']]
    )
    fig.update_xaxes(tickfont=dict(size=10))
    yaxis_names = ['yaxis'] + [axis_name for axis_name in fig.layout._subplotid_props if 'yaxis' in axis_name]
    yaxis_layout_dict = {yaxis_name + "_tickformat": ',.2%' for yaxis_name in yaxis_names}
    fig.update_layout(yaxis_layout_dict)
    height = 640
    if 'facet_row' in config.keys():
        height = max(640, 300 * len(ih_analysis[config['facet_row']].unique()))

    fig.update_layout(
        xaxis_title=x_axis,
        yaxis_title=y_axis,
        hovermode="x unified",
        height=height
    )
    fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
    fig = fig.update_traces(hovertemplate=x_axis + ' : %{x}' + '<br>' +
                                          config['color'] + ' : %{customdata[0]}' + '<br>' +
                                          y_axis + ' : %{y:.2%}' + '<extra></extra>')

    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    return ih_analysis


@timed
def model_ml_scores_line_plot_roc_pr_curve(data: Union[pl.DataFrame, pd.DataFrame],
                                           config: dict) -> pd.DataFrame:
    y_axis = config.get('y', None)
    if y_axis == "roc_auc":
        x = 'fpr'
        y = 'tpr'
        title = config['description'] + ": ROC Curve"
        label_x = 'False Positive Rate'
        label_y = 'True Positive Rate'
        x0 = 0
        y0 = 0
        x1 = 1
        y1 = 1
    elif y_axis == "average_precision":
        x = 'recall'
        y = 'precision'
        title = config['description'] + ": Precision-Recall Curve Curve"
        label_x = 'Recall'
        label_y = 'Precision'
        x0 = 0
        y0 = 1
        x1 = 1
        y1 = 0
    else:
        ih_analysis = model_ml_scores_line_plot(data, config)
        return ih_analysis

    pills1, toggle1 = st.columns(2)
    options = ["Curves", "Calibration", "Gain", "Lift"]
    selection = pills1.pills("Additional plots", options, selection_mode="single")
    # curves_on = toggle1.toggle("Show as curves", value=False, help="Show as curve (ROC or PR).",
    #                           key="Curves" + config['description'])
    # calibration_on = toggle3.toggle("Calibration plot", value=False, help="Show calibration plot.",
    #                                key="Calibration" + config['description'])
    adv_on = toggle1.toggle("Advanced options", value=False, key="Advanced options" + config['description'],
                            help="Show advanced reporting options")

    # if curves_on and calibration_on:
    #    st.warning('Select either curves or calibration.')
    #    st.stop()

    xplot_y_bool = False
    xplot_col = config.get('color', None)
    facet_row = '---' if not 'facet_row' in config.keys() else config['facet_row']
    facet_column = '---' if not 'facet_column' in config.keys() else config['facet_column']
    x_axis = config.get('x', None)
    y_axis = config.get('y', None)
    property = PROPENSITY

    if adv_on:
        plot_menu = get_plot_parameters_menu_ml(config=config, is_y_axis_required=False)
        x_axis = plot_menu['x']
        y_axis = plot_menu['y']
        facet_column = plot_menu['facet_col']
        facet_row = plot_menu['facet_row']
        xplot_col = plot_menu['color']
        property = plot_menu['property']

    grp_by = [x_axis]
    if not (facet_column == '---'):
        if not facet_column in grp_by:
            grp_by.append(facet_column)
    else:
        facet_column = None
    if not (facet_row == '---'):
        if not facet_row in grp_by:
            grp_by.append(facet_row)
    else:
        facet_row = None

    if not (xplot_col == '---'):
        if not xplot_col in grp_by:
            grp_by.append(xplot_col)
    else:
        xplot_col = None

    cp_config = config.copy()
    cp_config['x'] = x_axis
    cp_config['group_by'] = grp_by
    cp_config['color'] = xplot_col
    cp_config['facet_row'] = facet_row
    cp_config['facet_column'] = facet_column
    cp_config['log_y'] = xplot_y_bool
    cp_config['property'] = property

    ih_analysis = pd.DataFrame()
    if selection == 'Curves':
        report_data = data.copy()
        report_data = filter_dataframe(align_column_types(report_data), case=False)
        cp_config = config.copy()
        cp_config['group_by'] = list(set(([facet_row] if facet_row is not None else []) + (
            [facet_column] if facet_column is not None else []) + (
                                             [xplot_col] if xplot_col is not None else [])))
        if not cp_config['group_by']:
            cp_config['group_by'] = None
        report_data = calculate_model_ml_scores(report_data, cp_config, False)
        if cp_config['group_by'] is None:
            report_data = report_data.with_columns(
                [
                    pl.col(x).list.first().alias(x),
                    pl.col(y).list.first().alias(y)
                ]
            )
        report_data = report_data.explode([x, y])
        fig = px.line(report_data,
                      x=x, y=y,
                      title=title,
                      color=xplot_col,
                      log_y=xplot_y_bool,
                      facet_col=facet_column,
                      facet_row=facet_row,
                      height=len(
                          report_data[facet_row].unique()) * 400 if facet_row is not None else 640
                      )
        if y_axis == "roc_auc":
            fig.add_shape(
                type="line", line=dict(dash='dash', color="darkred"),
                row='all', col='all', x0=x0, y0=y0, x1=x1, y1=y1
            )
        fig.update_layout(
            xaxis=dict(
                range=[0, 1]
            ),
            yaxis=dict(
                range=[0, 1]
            )
        )
        fig.for_each_xaxis(lambda x: x.update({'title': ''}))
        fig.for_each_yaxis(lambda y: y.update({'title': ''}))
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
        fig.add_annotation(
            showarrow=False,
            xanchor='center',
            xref='paper',
            x=0.5,
            yref='paper',
            y=-0.1,
            text=label_x
        )
        fig.add_annotation(
            showarrow=False,
            xanchor='center',
            xref='paper',
            x=-0.04,
            yanchor='middle',
            yref='paper',
            y=0.5,
            textangle=90,
            text=label_y
        )

        st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    elif selection == 'Calibration':
        x = 'calibration_proba'
        y = 'calibration_rate'
        title = config['description'] + ": Calibration Plot"
        label_x = 'Probabilities'
        label_y = 'Positives share'
        x0 = 0
        y0 = 0
        x1 = 1
        y1 = 1

        report_data = data.copy()
        report_data = filter_dataframe(align_column_types(report_data), case=False)
        cp_config = config.copy()
        cp_config['group_by'] = list(set(([facet_row] if facet_row is not None else []) + (
            [facet_column] if facet_column is not None else []) + (
                                             [xplot_col] if xplot_col is not None else [])))
        if not cp_config['group_by']:
            cp_config['group_by'] = None
        report_data = calculate_model_ml_scores(report_data, cp_config, False)
        if cp_config['group_by'] is None:
            report_data = (report_data
                           .with_columns([pl.col(x).list.first().alias(x), pl.col(y).list.first().alias(y)])
                           .sort(x, descending=False))
        report_data = report_data.explode([x, y]).sort(x, descending=False)
        fig = px.line(report_data,
                      x=x, y=y,
                      title=title,
                      color=xplot_col,
                      log_y=xplot_y_bool,
                      facet_col=facet_column,
                      facet_row=facet_row,
                      height=len(
                          report_data[facet_row].unique()) * 400 if facet_row is not None else 640
                      )
        fig.add_shape(
            type="line", line=dict(dash='dash', color="darkred"),
            row='all', col='all', x0=x0, y0=y0, x1=x1, y1=y1
        )
        fig.update_layout(
            xaxis=dict(
                range=[0, 1]
            ),
            yaxis=dict(
                range=[0, 1]
            )
        )
        fig.for_each_xaxis(lambda x: x.update({'title': ''}))
        fig.for_each_yaxis(lambda y: y.update({'title': ''}))
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
        fig.add_annotation(
            showarrow=False,
            xanchor='center',
            xref='paper',
            x=0.5,
            yref='paper',
            y=-0.1,
            text=label_x
        )
        fig.add_annotation(
            showarrow=False,
            xanchor='center',
            xref='paper',
            x=-0.04,
            yanchor='middle',
            yref='paper',
            y=0.5,
            textangle=90,
            text=label_y
        )

        st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    elif selection == 'Gain':
        x = 'sample_fraction'
        y = 'gain'
        title = config['description'] + ": Gain Plot"
        label_x = 'Fraction of population'
        label_y = 'Gain'
        x0 = 0
        y0 = 0
        x1 = 1
        y1 = 1

        report_data = data.copy()
        report_data = filter_dataframe(align_column_types(report_data), case=False)
        cp_config = config.copy()
        cp_config['group_by'] = list(set(([facet_row] if facet_row is not None else []) + (
            [facet_column] if facet_column is not None else []) + (
                                             [xplot_col] if xplot_col is not None else [])))
        if not cp_config['group_by']:
            cp_config['group_by'] = None
        report_data = calculate_model_ml_scores(report_data, cp_config, False)
        if cp_config['group_by'] is None:
            report_data = (report_data
                           .with_columns([pl.col(x).list.first().alias(x), pl.col(y).list.first().alias(y)])
                           .sort(x, descending=False))
        list_cols = ["tpr", "fpr"]
        report_data = (
            report_data.explode(list_cols)
            .drop("calibration_rate", strict=False)
            .with_columns([
                (pl.col("pos_fraction") * pl.col("tpr") + (1.0 - pl.col("pos_fraction")) * pl.col("fpr"))
                .alias("sample_fraction"),
                pl.col("tpr").alias("gain"),
            ])
        )
        fig = px.line(report_data,
                      x=x, y=y,
                      title=title,
                      color=xplot_col,
                      facet_col=facet_column,
                      facet_row=facet_row,
                      height=len(
                          report_data[facet_row].unique()) * 400 if facet_row is not None else 640
                      )
        fig.add_shape(
            type="line", line=dict(dash='dash', color="darkred"),
            row='all', col='all', x0=x0, y0=y0, x1=x1, y1=y1
        )
        fig.update_layout(
            xaxis=dict(
                range=[0, 1]
            ),
            yaxis=dict(
                range=[0, 1]
            )
        )
        fig.for_each_xaxis(lambda x: x.update({'title': ''}))
        fig.for_each_yaxis(lambda y: y.update({'title': ''}))
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
        fig.add_annotation(
            showarrow=False,
            xanchor='center',
            xref='paper',
            x=0.5,
            yref='paper',
            y=-0.1,
            text=label_x
        )
        fig.add_annotation(
            showarrow=False,
            xanchor='center',
            xref='paper',
            x=-0.03,
            yanchor='middle',
            yref='paper',
            y=0.5,
            textangle=90,
            text=label_y
        )

        st.plotly_chart(fig, use_container_width=True, theme="streamlit")
        ih_analysis = report_data.select(
            (cp_config['group_by'] if cp_config['group_by'] else []) + ['pos_fraction', 'sample_fraction',
                                                                        'gain']).to_pandas()
    elif selection == 'Lift':
        x = 'sample_fraction'
        y = 'lift'
        title = config['description'] + ": Lift Plot"
        label_x = 'Fraction of population'
        label_y = 'Lift'
        x0 = 0
        y0 = 1
        x1 = 1
        y1 = 1

        report_data = data.copy()
        report_data = filter_dataframe(align_column_types(report_data), case=False)
        cp_config = config.copy()
        cp_config['group_by'] = list(set(([facet_row] if facet_row is not None else []) + (
            [facet_column] if facet_column is not None else []) + (
                                             [xplot_col] if xplot_col is not None else [])))
        if not cp_config['group_by']:
            cp_config['group_by'] = None
        report_data = calculate_model_ml_scores(report_data, cp_config, False)
        if cp_config['group_by'] is None:
            report_data = (report_data
                           .with_columns([pl.col(x).list.first().alias(x), pl.col(y).list.first().alias(y)])
                           .sort(x, descending=False))
        list_cols = ["tpr", "fpr"]
        report_data = (
            report_data.explode(list_cols)
            .drop("calibration_rate", strict=False)
            .with_columns([
                (pl.col("pos_fraction") * pl.col("tpr") + (1.0 - pl.col("pos_fraction")) * pl.col("fpr"))
                .alias("sample_fraction"),
                pl.col("tpr").alias("gain"),
                (pl.col("tpr") / (
                        pl.col("pos_fraction") * pl.col("tpr") + (1.0 - pl.col("pos_fraction")) * pl.col("fpr")))
                .alias("lift"),
            ])
            .with_columns([
                pl.when(pl.col("sample_fraction") > 0.000001)
                .then(pl.col("gain") / pl.col("sample_fraction"))
                .otherwise(0)
                .alias("lift")
            ])
        )
        fig = px.line(report_data,
                      x=x, y=y,
                      title=title,
                      color=xplot_col,
                      facet_col=facet_column,
                      facet_row=facet_row,
                      height=len(
                          report_data[facet_row].unique()) * 400 if facet_row is not None else 640
                      )
        fig.add_shape(
            type="line", line=dict(dash='dash', color="darkred"),
            row='all', col='all', x0=x0, y0=y0, x1=x1, y1=y1
        )
        fig.update_layout(
            xaxis=dict(
                range=[0, 1]
            ),
            yaxis=dict(
                range=[0, report_data[y].max()]
            )
        )
        fig.for_each_xaxis(lambda x: x.update({'title': ''}))
        fig.for_each_yaxis(lambda y: y.update({'title': ''}))
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
        fig.add_annotation(
            showarrow=False,
            xanchor='center',
            xref='paper',
            x=0.5,
            yref='paper',
            y=-0.1,
            text=label_x
        )
        fig.add_annotation(
            showarrow=False,
            xanchor='center',
            xref='paper',
            x=-0.03,
            yanchor='middle',
            yref='paper',
            y=0.5,
            textangle=90,
            text=label_y
        )

        st.plotly_chart(fig, use_container_width=True, theme="streamlit")
        ih_analysis = report_data.select(
            (cp_config['group_by'] if cp_config['group_by'] else []) + ['pos_fraction', 'sample_fraction', 'gain',
                                                                        'lift']).to_pandas()
    else:
        ih_analysis = model_ml_scores_line_plot(data, cp_config)
    return ih_analysis


@timed
def model_ml_treemap_plot(data: Union[pl.DataFrame, pd.DataFrame],
                          config: dict) -> pd.DataFrame:
    report_data = calculate_reports_data(data, config).to_pandas()
    ih_analysis = filter_dataframe(align_column_types(report_data), case=False)
    if ih_analysis.shape[0] == 0:
        st.warning("No data available.")
        st.stop()
    fig = px.treemap(ih_analysis, path=[px.Constant("ALL")] + config['group_by'], values='Count',
                     color=config['color'],
                     color_continuous_scale=px.colors.sequential.RdBu_r,
                     title=config['description'],
                     height=640,
                     )
    fig.update_traces(textinfo="label+value+percent parent+percent root")
    fig.update_layout(margin=dict(t=50, l=25, r=25, b=25))
    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    return ih_analysis


def get_plot_parameters_menu_ml(config: dict, is_y_axis_required: bool = True):
    metric = config["metric"]
    m_config = get_config()["metrics"][metric]
    report_grp_by = m_config['group_by']
    report_grp_by = sorted(report_grp_by)
    scores = m_config['scores']

    xplot_col = config.get('color', None)
    facet_row = '---' if not 'facet_row' in config.keys() else config['facet_row']
    facet_column = '---' if not 'facet_column' in config.keys() else config['facet_column']
    x_axis = config.get('x', None)
    y_axis = config.get('y', None)
    property = PROPENSITY

    cols = st.columns(6 if is_y_axis_required else 5)
    with cols[0]:
        x_axis = st.selectbox(
            label='X-Axis',
            options=report_grp_by,
            index=report_grp_by.index(x_axis) if x_axis else 0,
            help="Select X-Axis."
        )
    with cols[1]:
        xplot_col = st.selectbox(
            label='Colour By',
            options=report_grp_by,
            index=report_grp_by.index(xplot_col) if xplot_col else 0,
            help="Select color."
        )
    with cols[2]:
        options_row = ['---'] + report_grp_by
        if 'facet_row' in config.keys():
            facet_row = st.selectbox(
                label='Row Facets',
                options=options_row,
                index=options_row.index(config['facet_row']),
                help="Select data column."
            )
        else:
            facet_row = st.selectbox(
                label='Row Facets',
                options=options_row,
                help="Select data column."
            )
    with cols[3]:
        options_col = ['---'] + report_grp_by
        if 'facet_column' in config.keys():
            facet_column = st.selectbox(
                label='Column Facets',
                options=options_col,
                index=options_col.index(config['facet_column']),
                help="Select data column."
            )
        else:
            facet_column = st.selectbox(
                label='Column Facets',
                options=options_col,
                help="Select data column."
            )
    if is_y_axis_required:
        with cols[4]:
            y_axis = st.selectbox(
                label='Y-Axis',
                options=scores,
                index=scores.index(y_axis) if y_axis else 0,
                help="Select Y-Axis."
            )
    with cols[len(cols) - 1]:
        property = st.selectbox(
            label='Property',
            options=[PROPENSITY, FINAL_PROPENSITY],
            help="Select Property."
        )

    return {'x': x_axis, 'color': xplot_col, 'facet_row': facet_row, 'facet_col': facet_column, 'y': y_axis,
            'property': property}


def ml_scores_card(ih_analysis: Union[pl.DataFrame, pd.DataFrame], metric_name: str):
    config = dict()
    config['metric'] = metric_name
    df = calculate_model_ml_scores(ih_analysis, config, True)
    st.metric(label="**Model AUC**", value='{:.2%}'.format(df["roc_auc"].item()), border=False,
              delta=f"Avg Precision = {'{:.2%}'.format(df['average_precision'].item())}", delta_color='off',
              help=f'Model ROC AUC and Average Precision')


def ml_scores_pers_card(ih_analysis: Union[pl.DataFrame, pd.DataFrame], metric_name: str):
    config = dict()
    config['metric'] = metric_name
    df = calculate_model_ml_scores(ih_analysis, config, True)
    st.metric(label="**Personalization**", value='{:.2}'.format(df["personalization"].item()), border=False,
              delta=f"Novelty = {'{:.2}'.format(df['novelty'].item())}", delta_color='off',
              help=f'Personalization and Novelty')



import traceback
from typing import List, Sequence

import datasketches
import numpy as np
import polars as pl
import polars_ds as pds
from polars import Series
from polars import selectors as cs
from polars_ds import weighted_mean
from sklearn.feature_extraction import FeatureHasher
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import normalize

from value_dashboard.metrics.constants import NAME, CUSTOMER_ID, INTERACTION_ID, RANK, OUTCOME, PROPENSITY, \
    FINAL_PROPENSITY
from value_dashboard.utils.config import get_config
from value_dashboard.utils.polars_utils import merge_digests, build_digest
from value_dashboard.utils.string_utils import strtobool
from value_dashboard.utils.timer import timed


def personalization(args: List[Series]) -> pl.Float64:
    """
    Personalization measures recommendation similarity across users.
    A high score indicates good personalization (user's lists of recommendations are different).
    A low score indicates poor personalization (user's lists of recommendations are very similar).
    A model is "personalizing" well if the set of recommendations for each user is different.
    Inspired by implementation at https://github.com/statisticianinstilettos/recmetrics.git (unfortunately could not use it in polars context)
    Parameters:
    ----------
    args : a 2-elements list of Series with CustomerID and Action Name
    Returns:
    -------
        The personalization score for all recommendations.
    """
    if len(args) < 2:
        return 0.0
    df = pl.DataFrame(args, schema=[CUSTOMER_ID, NAME])
    height = df.height
    if height < 5000:
        pass
    elif height < 10000:
        df = df.slice(round(height / 2))
    else:
        df = df.slice(round(height / 2), 5000)
    h = FeatureHasher(input_type="string", n_features=2 ** 16)
    df = (
        df
        .group_by([CUSTOMER_ID])
        .agg([
            pl.col(NAME).alias("ActionNames")
        ])
    )

    predicted = df.get_column("ActionNames").to_list()
    rec_matrix_sparse = h.transform(predicted)
    similarity = cosine_similarity(X=rec_matrix_sparse, dense_output=False)
    dim = similarity.shape[0]
    if dim == 1:
        return 0.0
    personalization_score = (similarity.sum() - dim) / (dim * (dim - 1))
    return 1 - personalization_score


def personalization_optimized(args: List[Series]) -> pl.Float64:
    """
    Personalization measures recommendation similarity across users (Optimized).
    Avoids computing the full N x N similarity matrix.

    Parameters:
    ----------
    args : a 2-elements list of Series with CustomerID and Action Name

    Returns:
    -------
    The personalization score for all recommendations.
    """
    if len(args) < 2:
        return 0.0

    df = pl.DataFrame(args, schema=[CUSTOMER_ID, NAME])
    if df.height < 2:
        return 0.0
    df = pl.DataFrame(args, schema=[CUSTOMER_ID, NAME])
    height = df.height
    if height < 50000:
        pass
    elif height < 100000:
        df = df.slice(round(height / 2))
    else:
        df = df.slice(round(height / 2), 50000)

    df_grouped = (
        df
        .group_by(CUSTOMER_ID)
        .agg(pl.col(NAME).alias("ActionNames"))
    )

    predicted = df_grouped.get_column("ActionNames").to_list()
    dim = len(predicted)

    if dim <= 1:
        return 0.0

    h = FeatureHasher(input_type="string", n_features=2 ** 16)
    rec_matrix_sparse = h.transform(predicted)
    rec_matrix_normalized = normalize(rec_matrix_sparse, norm='l2', axis=1)
    total_similarity_sum = np.sum(np.square(rec_matrix_normalized.sum(axis=0)))
    off_diagonal_similarity_sum = total_similarity_sum - dim
    if dim > 1:
        avg_off_diagonal_similarity = off_diagonal_similarity_sum / (dim * (dim - 1))
    else:
        avg_off_diagonal_similarity = 0.0

    personalization_score = 1.0 - avg_off_diagonal_similarity

    return personalization_score


def novelty(args: Sequence[Series]) -> float:
    if len(args) < 2:
        return 0.0
    df = pl.DataFrame(args, schema=[CUSTOMER_ID, INTERACTION_ID, NAME])
    height = df.height
    if height < 50000:
        pass
    elif height < 100000:
        df = df.slice(round(height / 2))
    else:
        df = df.slice(round(height / 2), 50000)
    u = df.n_unique(subset=[CUSTOMER_ID])
    if u == 0:
        return 0.0
    item_counts = (
        df.group_by(NAME)
        .agg(pl.len().alias("ActionCount"))
        .with_columns(
            (pl.col("ActionCount") * -((pl.col("ActionCount") / u).log(base=2) + 1e-10)).alias("total_self_info")
        )
    )
    total_self_info = item_counts["total_self_info"].sum()
    n = (
        df.group_by(INTERACTION_ID)
        .agg(pl.len().alias("RecLength"))
        ["RecLength"].max()
    )
    if n == 0:
        return 0.0
    novelty_score = total_self_info / (u * n)
    return float(novelty_score)


def binary_metrics_tdigest(args: List[Series]) -> pl.Struct:
    pos_series, neg_series = args[0], args[1]
    if pos_series.len() == 0 or neg_series.len() == 0:
        return {
            'roc_auc': 0.0,
            'average_precision': 0.0,
            'tpr': [0.0],
            'fpr': [0.0],
            'precision': [0.0],
            'recall': [0.0],
            'pos_fraction': 0.0
        }

    thresholds = [round(float(t), 4) for t in np.linspace(0, 1, num=101)]

    all_pos = args[0]
    all_neg = args[1]

    pos_sk = datasketches.tdigest_double.deserialize(merge_digests([all_pos]))
    neg_sk = datasketches.tdigest_double.deserialize(merge_digests([all_neg]))

    if pos_sk.is_empty() or neg_sk.is_empty():
        return {
            'roc_auc': 0.0,
            'average_precision': 0.0,
            'tpr': [0.0],
            'fpr': [0.0],
            'precision': [0.0],
            'recall': [0.0],
            'pos_fraction': 0.0
        }

    pos_count = pos_sk.get_total_weight()
    neg_count = neg_sk.get_total_weight()

    cdf_pos = np.array(pos_sk.get_cdf(thresholds))
    cdf_neg = np.array(neg_sk.get_cdf(thresholds))
    tpr = 1.0 - cdf_pos
    fpr = 1.0 - cdf_neg

    idx = np.argsort(fpr)
    fpr_sorted = fpr[idx]
    tpr_sorted = tpr[idx]
    roc_auc = np.trapz(tpr_sorted, fpr_sorted)

    recall = tpr[::-1]
    fpr_desc = fpr[::-1]
    tp = pos_count * recall
    fp = neg_count * fpr_desc
    pos_fraction = pos_count / (pos_count + neg_count)

    eps = 1e-10
    precision = tp / (tp + fp + eps)

    precision = np.maximum.accumulate(precision[::-1])[::-1]

    if recall[0] != 0.0:
        recall = np.concatenate(([0.0], recall))
        precision = np.concatenate(([1.0], precision))

    delta_r = recall[1:] - recall[:-1]
    average_precision = np.sum(delta_r * precision[1:])

    return {
        'roc_auc': float(roc_auc),
        'average_precision': float(average_precision),
        'tpr': tpr_sorted.tolist(),
        'fpr': fpr_sorted.tolist(),
        'precision': precision.tolist(),
        'recall': recall.tolist(),
        'pos_fraction': pos_fraction
    }


def calibration_tdigest(args: List[Series]) -> pl.Struct:
    pos_series, neg_series = args[0], args[1]
    if pos_series.len() == 0 or neg_series.len() == 0:
        return {'calibration_bin': [0.0], 'calibration_proba': [0.0], 'calibration_rate': [0.0]}

    thresholds = [round(float(t), 4) for t in np.linspace(0, 0.1, num=10, endpoint=False)] + [round(float(t), 4) for t
                                                                                              in np.linspace(0.1, 1,
                                                                                                             num=17)]

    all_pos = args[0]
    all_neg = args[1]

    pos_sk = datasketches.tdigest_double.deserialize(merge_digests([all_pos]))
    neg_sk = datasketches.tdigest_double.deserialize(merge_digests([all_neg]))

    if pos_sk.is_empty() or neg_sk.is_empty():
        return {'calibration_bin': [0.0], 'calibration_proba': [0.0], 'calibration_rate': [0.0]}

    pos_count = pos_sk.get_total_weight()
    neg_count = neg_sk.get_total_weight()
    cdf_pos = np.array(pos_sk.get_cdf(thresholds))
    cdf_neg = np.array(neg_sk.get_cdf(thresholds))
    delta_pos = cdf_pos[1:] - cdf_pos[:-1]
    delta_neg = cdf_neg[1:] - cdf_neg[:-1]

    pos_in_bin = pos_count * delta_pos
    neg_in_bin = neg_count * delta_neg
    total_in_bin = pos_in_bin + neg_in_bin
    positives_rate = pos_in_bin / total_in_bin

    n_quantiles = 10
    results_bin = []
    results_proba = []
    results_rate = []

    for i in range(len(thresholds) - 1):
        a, b = thresholds[i], thresholds[i + 1]
        cdf_a_pos = pos_sk.get_cdf([a])[0]
        cdf_b_pos = pos_sk.get_cdf([b])[0]
        cdf_a_neg = neg_sk.get_cdf([a])[0]
        cdf_b_neg = neg_sk.get_cdf([b])[0]

        if (cdf_b_pos - cdf_a_pos) == 0 and (cdf_b_neg - cdf_a_neg) == 0:
            mean_propensity = (a + b) / 2
        else:
            q_pos = np.linspace(cdf_a_pos, cdf_b_pos, n_quantiles, endpoint=False)
            if q_pos.size > 0:
                pos_vals = [pos_sk.get_quantile(q) for q in q_pos]
                pos_bin_mean = np.mean(pos_vals)
            else:
                pos_bin_mean = 0.0

            q_neg = np.linspace(cdf_a_neg, cdf_b_neg, n_quantiles, endpoint=False)
            if q_neg.size > 0:
                neg_vals = [neg_sk.get_quantile(q) for q in q_neg]
                neg_bin_mean = np.mean(neg_vals)
            else:
                neg_bin_mean = 0.0

            total = pos_in_bin[i] + neg_in_bin[i]
            if total > 0:
                mean_propensity = (pos_bin_mean * pos_in_bin[i] + neg_bin_mean * neg_in_bin[i]) / total
            else:
                mean_propensity = (a + b) / 2

        results_bin.append((a + b) / 2)
        results_proba.append(float(mean_propensity))
        results_rate.append(float(positives_rate[i]) if total_in_bin[i] > 0 else 0.0)

    return {'calibration_bin': results_bin, 'calibration_proba': results_proba, 'calibration_rate': results_rate}


@timed
def model_ml_scores(ih: pl.LazyFrame, config: dict, streaming=False, background=False):
    grp_by = list(set(config['group_by'] + get_config()["metrics"]["global_filters"]))
    negative_model_response = config['negative_model_response']
    positive_model_response = config['positive_model_response']
    use_t_digest = strtobool(config['use_t_digest']) if 'use_t_digest' in config.keys() else True

    if "filter" in config:
        filter_exp_cmp = config["filter"]
        if not isinstance(filter_exp_cmp, str):
            ih = ih.filter(config["filter"])

    common_aggs = [
        pl.len().alias('Count'),
        pl.map_groups(
            exprs=[CUSTOMER_ID, NAME],
            function=personalization_optimized,
            return_dtype=pl.Float64,
            returns_scalar=True
        ).alias("personalization"),
        pl.map_groups(
            exprs=[CUSTOMER_ID, INTERACTION_ID, NAME],
            function=novelty,
            return_dtype=pl.Float64,
            returns_scalar=True
        ).alias("novelty")
    ]
    if use_t_digest:
        t_digest_aggs = [
            pl.map_groups(
                exprs=[pl.col(PROPENSITY)
                       .filter(pl.col("Outcome_Boolean") == True)],
                function=lambda s: build_digest(s),
                return_dtype=pl.Binary,
                returns_scalar=True
            ).alias('tdigest_positives'),
            pl.map_groups(
                exprs=[pl.col(PROPENSITY)
                       .filter(pl.col("Outcome_Boolean") == False)],
                function=lambda s: build_digest(s),
                return_dtype=pl.Binary,
                returns_scalar=True
            ).alias('tdigest_negatives'),
            pl.map_groups(
                exprs=[pl.col(FINAL_PROPENSITY)
                       .filter(pl.col("Outcome_Boolean") == True)],
                function=lambda s: build_digest(s),
                return_dtype=pl.Binary,
                returns_scalar=True
            ).alias('tdigest_finalprop_positives'),
            pl.map_groups(
                exprs=[pl.col(FINAL_PROPENSITY)
                       .filter(pl.col("Outcome_Boolean") == False)],
                function=lambda s: build_digest(s),
                return_dtype=pl.Binary,
                returns_scalar=True
            ).alias('tdigest_finalprop_negatives')
        ]
        agg_exprs = common_aggs + t_digest_aggs
    else:
        extra_aggs = [pds.query_binary_metrics(
            "Outcome_Boolean",
            PROPENSITY,
            threshold=0).alias("metrics")]
        agg_exprs = common_aggs + extra_aggs

    try:
        ml_data = (
            ih
            .filter(
                (pl.col(OUTCOME).is_in(negative_model_response + positive_model_response))
            )
            .with_columns([
                pl.when(pl.col(OUTCOME).is_in(positive_model_response)).
                then(True).otherwise(False).alias('Outcome_Boolean')
            ])
            .filter(pl.any("Outcome_Boolean").over(grp_by))
            .filter(pl.col('Outcome_Boolean') == pl.col('Outcome_Boolean').max().over(INTERACTION_ID, NAME, RANK))
            .group_by(grp_by)
            .agg(agg_exprs)
        )
        if not use_t_digest:
            ml_data = ml_data.unnest(pl.col("metrics"))

        if background:
            return ml_data.collect(background=background, engine="streaming" if streaming else "auto")
        else:
            return ml_data
    except Exception as e:
        print("An exception occurred: ", e)
        traceback.print_exc()
        raise e


@timed
def compact_model_ml_scores_data(model_roc_auc_data: pl.DataFrame,
                                 config: dict) -> pl.DataFrame:
    auc_data = model_roc_auc_data.filter(pl.col("Count") > 0)
    use_t_digest = strtobool(config['use_t_digest']) if 'use_t_digest' in config.keys() else True

    grp_by = config['group_by'] + get_config()["metrics"]["global_filters"]
    scores = config["scores"]
    tdigest_columns = [col for col in auc_data.collect_schema().names() if col.startswith("tdigest")]
    grp_by = list(set(grp_by))
    auc_data = (
        auc_data
        .group_by(grp_by)
        .agg(
            (
                [
                    weighted_mean(pl.col(scores), pl.col("Count")).name.suffix("_a")
                ]
                if not use_t_digest else
                [
                    weighted_mean(pl.col(['personalization', 'novelty']), pl.col("Count")).name.suffix("_a")
                ]
            )
            +
            [
                pl.col("Count").sum().alias("Count_a"),
                pl.col(grp_by).first().name.suffix("_a")
            ]
            +
            (
                [
                    pl.map_groups(
                        exprs=[pl.col(f'{c}')],
                        function=lambda s: merge_digests(s),
                        return_dtype=pl.Binary,
                        returns_scalar=True
                    ).alias(f'{c}_a') for c in tdigest_columns
                ] if use_t_digest else []
            )
        )
        .select(cs.ends_with("_a"))
        .rename(lambda column_name: column_name.removesuffix('_a'))
    )

    return auc_data



from functools import lru_cache

import plotly.graph_objects as go
import plotly.io as pio


@lru_cache(maxsize=1)
def init_plotly_theme():
    adjusted_colors = [
        "#3498db", "#2ecc71", "#e74c3c", "#9b59b6",
        "#f39c12", "#d35400", "#1abc9c", "red", "green", "blue", "#7b3f00",
        "purple", "yellow", "black", "darkblue", "darkgreen", "darkred", "#4b006e",
        "steelblue", "lightgreen", "lightblue", "#fffacd"
    ]

    base = pio.templates.default  # pio.templates["plotly_white"]
    new_template = pio.templates[base].update(layout=go.Layout(colorway=adjusted_colors))
    pio.templates["cdhvd"] = new_template
    pio.templates.default = "cdhvd"



import traceback

import polars as pl

from value_dashboard.metrics.constants import INTERACTION_ID, NAME, OUTCOME, RANK
from value_dashboard.utils.config import get_config
from value_dashboard.utils.timer import timed


@timed
def experiment(ih: pl.LazyFrame, config: dict, streaming=False, background=False):
    mand_props_grp_by = list(set(get_config()["metrics"]["global_filters"] +
                                 config['group_by'] + [config['experiment_name']] + [config['experiment_group']]))
    negative_model_response = config['negative_model_response']
    positive_model_response = config['positive_model_response']

    if "filter" in config:
        filter_exp_cmp = config["filter"]
        if not isinstance(filter_exp_cmp, str):
            ih = ih.filter(config["filter"])

    try:
        ih_analysis = (
            ih.filter(
                (pl.col(OUTCOME).is_in(negative_model_response + positive_model_response))
            )
            .with_columns([
                pl.when(pl.col(OUTCOME).is_in(positive_model_response)).
                then(1).otherwise(0).alias('Outcome_Binary')
            ])
            .filter(pl.col('Outcome_Binary') == pl.col('Outcome_Binary').max().over(INTERACTION_ID, NAME, RANK))
            .group_by(mand_props_grp_by)
            .agg([
                pl.len().alias('Count'),
                pl.sum("Outcome_Binary").alias("Positives")
            ])
            .filter(pl.col("Count") > 0)
            .with_columns([
                (pl.col("Count") - (pl.col("Positives"))).alias("Negatives")
            ])
        )
        if background:
            return ih_analysis.collect(background=background, engine="streaming" if streaming else "auto")
        else:
            return ih_analysis
    except Exception as e:
        traceback.print_exc()
        raise e


@timed
def compact_experiment_data(exp_data: pl.DataFrame,
                            config: dict) -> pl.DataFrame:
    grp_by = config['group_by'] + [config['experiment_name']] + get_config()["metrics"]["global_filters"]
    grp_by = list(set(grp_by))
    if grp_by:
        exp_data = (
            exp_data
            .group_by(list(set(grp_by + [config['experiment_group']])))
            .agg([
                pl.col("Count").sum(),
                pl.col("Positives").sum(),
                pl.col("Negatives").sum(),
            ])
            .filter(
                (pl.col("Positives") > 0) & (pl.col("Negatives") > 0)
            )
        )

    return exp_data



import logging
from typing import Dict, Any, List

import datasketches
import numpy as np
import polars as pl
from polars.datatypes._parse import NoneType

from value_dashboard.utils.logger import get_logger

T_DIGEST_COMPRESSION = 500
REQ_SKETCH_ACCURACY = 24
logger = get_logger(__name__, logging.DEBUG)


def df_to_dict(df: pl.DataFrame, key_col: str, value_col: str) -> Dict[Any, Any]:
    """
    Get a Python dict from two columns of a DataFrame
    If the key column is not unique, the last row is used
    """
    return dict(df.select(key_col, value_col).iter_rows())


def schema_with_unique_counts(df: pl.DataFrame) -> pl.DataFrame:
    """
    Return a Polars DataFrame schema with the number of unique values for each string column.

    Parameters
    ----------
    df : pl.DataFrame
        The input Polars DataFrame.

    Returns
    -------
    pl.DataFrame
        A DataFrame with columns: 'column', 'dtype', and 'unique_count'.
    """

    schema = df.schema
    records = []
    for col, dtype in schema.items():
        if col.lower().endswith('id'):
            records.append({
                "Column": col,
                "Data Type": str(dtype),
                "Unique Count": "N/A",
                "Mode": "N/A",
                "Values": ''
            })
        elif dtype == pl.Utf8:
            unique_count = df[col].n_unique()
            mode = df[col].mode().to_list()
            unique = df[col].unique().to_list()
            records.append({
                "Column": col,
                "Data Type": str(dtype),
                "Unique Count": unique_count,
                "Mode": str(mode),
                "Values": str(unique) if unique_count < 10 else '...'
            })
        elif dtype == NoneType:
            unique_count = df[col].n_unique()
            mode = df[col].mode().to_list()
            unique = df[col].unique().to_list()
            records.append({
                "Column": col,
                "Data Type": str(dtype),
                "Unique Count": unique_count,
                "Mode": str(mode),
                "Values": str(unique) if unique_count < 10 else '...'
            })
        elif dtype.is_numeric():
            records.append({
                "Column": col,
                "Data Type": str(dtype),
                "Unique Count": "N/A",
                "Mode": "N/A",
                "Values": "Min = " + f'{df[col].min():.4f}' + " Max = " + f'{df[col].max():.4f}' + " Mean = "
                          + f'{df[col].mean():.4f}' + " Median = " + f'{df[col].median():.4f}'
            })
        else:
            records.append({
                "Column": col,
                "Data Type": str(dtype),
                "Unique Count": "N/A",
                "Mode": "N/A",
                "Values": "Min = " + f'{df[col].min()}' + " Max = " + f'{df[col].max()}'
            })

    return pl.DataFrame(records)


def build_digest(args: List[pl.Series]) -> bytes:
    arr = np.array(args[0].drop_nulls(), dtype=np.float64)
    sketch = datasketches.tdigest_double(k=T_DIGEST_COMPRESSION)
    sketch.update(arr)
    return sketch.serialize()


def merge_digests(args: List[pl.Series]
                  ) -> bytes:
    sketch_bytes_list = args[0].to_list()
    if not sketch_bytes_list:
        sketch = datasketches.tdigest_double(k=T_DIGEST_COMPRESSION)
        sketch.update(0.0)
        return sketch.serialize()
    merged = datasketches.tdigest_double.deserialize(sketch_bytes_list[0])
    for b in sketch_bytes_list[1:]:
        other = datasketches.tdigest_double.deserialize(b)
        merged.merge(other)
    return merged.serialize()


def estimate_quantile(args: List[pl.Series], quantile: float) -> float:
    sketch_bytes_list = args[0].to_list()
    if not sketch_bytes_list:
        return 0.0
    merged = datasketches.tdigest_double.deserialize(sketch_bytes_list[0])
    for b in sketch_bytes_list[1:]:
        other = datasketches.tdigest_double.deserialize(b)
        merged.merge(other)
    return merged.get_quantile(quantile)


def estimate_quantiles_arr(args: List[pl.Series], quantiles: List[float]) -> List[float]:
    sketch_bytes_list = args[0].to_list()
    if not sketch_bytes_list:
        return [0.0]
    merged = datasketches.tdigest_double.deserialize(sketch_bytes_list[0])
    for b in sketch_bytes_list[1:]:
        other = datasketches.tdigest_double.deserialize(b)
        merged.merge(other)
    return [merged.get_quantile(quantile) for quantile in quantiles]


def digest_to_histogram(tdigest: bytes, bins: int = 30, value_range: tuple[float, float] = None):
    tdigest = datasketches.tdigest_double.deserialize(tdigest)
    if value_range is None:
        value_range = (tdigest.get_quantile(0), tdigest.get_quantile(1))
    bin_edges = np.linspace(*value_range, bins + 1)
    bin_counts = []
    for left, right in zip(bin_edges[:-1], bin_edges[1:]):
        count = tdigest.get_cdf([right])[0] - tdigest.get_cdf([left])[0]
        bin_counts.append(count)
    return bin_edges, bin_counts



import traceback

import polars as pl

from value_dashboard.metrics.constants import MODELCONTROLGROUP, INTERACTION_ID, NAME, RANK, OUTCOME
from value_dashboard.utils.config import get_config
from value_dashboard.utils.timer import timed


@timed
def engagement(ih: pl.LazyFrame, config: dict, streaming=False, background=False):
    mand_props_grp_by = list(set(get_config()["metrics"]["global_filters"] + config['group_by'] + [MODELCONTROLGROUP]))
    negative_model_response = config['negative_model_response']
    positive_model_response = config['positive_model_response']

    if "filter" in config:
        filter_exp_cmp = config["filter"]
        if not isinstance(filter_exp_cmp, str):
            ih = ih.filter(config["filter"])

    try:
        ih_analysis = (
            ih.filter(
                (pl.col(OUTCOME).is_in(negative_model_response + positive_model_response))
            )
            .with_columns([
                pl.when(pl.col(OUTCOME).is_in(positive_model_response)).
                then(1).otherwise(0).alias('Outcome_Binary')
            ])
            .filter(pl.col('Outcome_Binary') == pl.col('Outcome_Binary').max().over(INTERACTION_ID, NAME, RANK))
            .group_by(mand_props_grp_by)
            .agg([
                pl.len().alias('Count'),
                pl.sum("Outcome_Binary").alias("Positives")
            ])
            .with_columns([
                (pl.col("Count") - (pl.col("Positives"))).alias("Negatives")
            ])
        )
        if background:
            return ih_analysis.collect(background=background, engine="streaming" if streaming else "auto")
        else:
            return ih_analysis
    except Exception as e:
        traceback.print_exc()
        raise e


@timed
def compact_engagement_data(eng_data: pl.DataFrame,
                            config: dict) -> pl.DataFrame:
    grp_by = config['group_by'] + get_config()["metrics"]["global_filters"]
    grp_by = list(set(grp_by))
    data_copy = (
        eng_data
        .group_by(grp_by + [MODELCONTROLGROUP])
        .agg(pl.sum("Negatives").alias("Negatives"),
             pl.sum("Positives").alias("Positives"),
             pl.sum("Count").alias("Count"))
    )

    return data_copy



import traceback

import polars as pl

from value_dashboard.metrics.constants import INTERACTION_ID, RANK, OUTCOME, CUSTOMER_ID, \
    CONVERSION_EVENT_ID, NAME
from value_dashboard.metrics.constants import REVENUE_PROP_NAME
from value_dashboard.utils.config import get_config
from value_dashboard.utils.timer import timed


@timed
def conversion(ih: pl.LazyFrame, config: dict, streaming=False, background=False):
    mand_props_grp_by = list(set(config['group_by'] + get_config()["metrics"]["global_filters"]))
    negative_model_response = config['negative_model_response']
    positive_model_response = config['positive_model_response']

    if "filter" in config:
        filter_exp_cmp = config["filter"]
        if not isinstance(filter_exp_cmp, str):
            ih = ih.filter(config["filter"])

    try:
        ih_analysis = ih.filter(
            (pl.col(OUTCOME).is_in(negative_model_response + positive_model_response))
        )
        ih_attribution = (
            ih_analysis
            .filter(pl.col(OUTCOME).is_in(positive_model_response))
            .group_by([CUSTOMER_ID, CONVERSION_EVENT_ID])
            .agg([
                pl.len().alias('Touchpoints')
            ]))
        ih_analysis = (
            ih_analysis
            .with_columns([
                pl.when(pl.col(OUTCOME).is_in(positive_model_response)).then(1).otherwise(0).alias('Outcome_Binary')
            ])
            .filter(pl.col('Outcome_Binary') == pl.col('Outcome_Binary').max().over(INTERACTION_ID, NAME, RANK))
            .join(ih_attribution, on=[CUSTOMER_ID, CONVERSION_EVENT_ID], how='left')
            .group_by(mand_props_grp_by)
            .agg([
                pl.len().alias('Count'),
                pl.sum(REVENUE_PROP_NAME),
                pl.sum('Touchpoints'),
                pl.sum("Outcome_Binary").alias("Positives")
            ])
            .with_columns([
                (pl.col("Count") - (pl.col("Positives"))).alias("Negatives")
            ])
        )
        if background:
            return ih_analysis.collect(background=background, engine="streaming" if streaming else "auto")
        else:
            return ih_analysis
    except Exception as e:
        traceback.print_exc()
        raise e


@timed
def compact_conversion_data(conv_data: pl.DataFrame,
                            config: dict) -> pl.DataFrame:
    data_copy = conv_data.filter(pl.col("Negatives") > 0)

    grp_by = list(set(config['group_by'] + get_config()["metrics"]["global_filters"]))
    if grp_by:
        data_copy = (
            data_copy
            .group_by(grp_by)
            .agg([
                pl.col("Negatives").sum(),
                pl.col("Positives").sum(),
                pl.col("Revenue").sum(),
                pl.col("Count").sum(),
                pl.col('Touchpoints').sum()
            ]
            )
        )

    return data_copy



import asyncio
import glob
import io
import json
import os
import re
import time
import typing
from collections import defaultdict
from datetime import timedelta
from pathlib import Path

import numpy as np
import polars as pl
import psutil
import streamlit as st
from polars import LazyFrame

from value_dashboard.datalake.df_db_proxy import PolarsDuckDBProxy
from value_dashboard.metrics.constants import INTERACTION_ID, RANK, OUTCOME, DROP_IH_COLUMNS, OUTCOME_TIME, \
    DECISION_TIME, ISSUE, GROUP, NAME, ACTION_ID
from value_dashboard.metrics.conversion import conversion
from value_dashboard.metrics.descriptive import descriptive
from value_dashboard.metrics.engagement import engagement
from value_dashboard.metrics.experiment import experiment
from value_dashboard.metrics.ml import model_ml_scores
from value_dashboard.pipeline.datatools import collect_ih_metrics_data
from value_dashboard.pipeline.datatools import compact_data
from value_dashboard.utils.config import get_config
from value_dashboard.utils.db_utils import save_file_meta, get_file_meta, drop_all_tables
from value_dashboard.utils.file_utils import read_dataset_export
from value_dashboard.utils.logger import get_logger
from value_dashboard.utils.string_utils import strtobool, capitalize
from value_dashboard.utils.timer import timed

IHFOLDER = "ihfolder"
logger = get_logger(__name__)
data_cache_hours = 24
if 'data_cache_hours' in get_config()['ux'].keys():
    data_cache_hours = get_config()['ux']['data_cache_hours']
logger.debug(f"Data will be cached for {data_cache_hours} hours.")
logger.debug(f"Numpy version {np.__version__}.")
logger.debug(f"Polars {pl.build_info()}.")
logger.debug(f"Polars threads: {pl.thread_pool_size()}.")


@timed
@st.cache_data(show_spinner=False, ttl=timedelta(hours=data_cache_hours))
def load_data() -> typing.Dict[str, pl.DataFrame]:
    load_start = time.time()
    if 'use_aggregated' in st.session_state:
        use_aggregated = st.session_state['use_aggregated']
        if use_aggregated:
            f = open(st.session_state['aggregated_path'], "r")
            aggregated = json.load(f)
            collected_metrics_data = {}
            for metric in aggregated.keys():
                frame = json.dumps(aggregated[metric])
                collected_metrics_data[metric] = pl.DataFrame.deserialize(source=io.StringIO(frame), format='json')
            f.close()
            st.session_state['drop_cache'] = False
            return collected_metrics_data

    drop_cache = st.session_state['drop_cache'] if 'drop_cache' in st.session_state else False

    db_proxy = PolarsDuckDBProxy()

    if drop_cache:
        drop_all_tables(db_proxy)

    processed_files = get_file_meta(db_proxy)

    if IHFOLDER not in st.session_state:
        st.warning("Please configure your files in the `Data import` tab.")
        st.stop()
    folder = st.session_state[IHFOLDER]
    if not os.path.isdir(folder):
        st.error(f"Folder {folder} not available anymore. Please update data import parameters.")
        st.stop()
    file_groups = defaultdict(set)
    config = get_config()
    filetype = config["ih"]["file_type"]
    logger.debug("File type: " + filetype)
    streaming = False
    if "streaming" in config['ih'].keys():
        streaming = strtobool(config['ih']['streaming'])
    logger.debug("Use polars streaming dataframe collect: " + str(streaming))
    if streaming:
        pl.Config.set_engine_affinity("streaming")
    background = False
    if "background" in config['ih'].keys():
        background = strtobool(config['ih']['background'])
    logger.debug("Use polars background dataframe collect: " + str(background))
    hive_partitioning = False
    if "hive_partitioning" in config['ih'].keys():
        hive_partitioning = strtobool(config['ih']['hive_partitioning'])
    logger.debug("Use hive partitioning: " + str(hive_partitioning))

    add_columns = config["ih"]["extensions"]["columns"]
    global_ih_filter = config["ih"]["extensions"]["filter"]

    if global_ih_filter:
        ih_filter_expr = eval(global_ih_filter)
    else:
        ih_filter_expr = pl.lit(True)

    if add_columns:
        add_columns_expr = eval(add_columns)
    else:
        add_columns_expr = []

    metrics = config["metrics"]
    for metric in metrics:
        params = metrics[metric]
        if isinstance(params, dict):
            if "filter" in params.keys():
                filter_exp_cmp = params["filter"]
                if isinstance(filter_exp_cmp, str):
                    if filter_exp_cmp:
                        params["filter"] = eval(filter_exp_cmp)

    metric_coroutines_map = {}
    for metric in metrics:
        if metric.startswith("engagement"):
            metric_coroutines_map[metric] = engagement
        if metric.startswith("model_ml_scores"):
            metric_coroutines_map[metric] = model_ml_scores
        if metric.startswith("conversion"):
            metric_coroutines_map[metric] = conversion
        if metric.startswith("descriptive"):
            metric_coroutines_map[metric] = descriptive
        if metric.startswith("experiment"):
            metric_coroutines_map[metric] = experiment

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    process = psutil.Process(os.getpid())

    # List all files in the folder matching the pattern
    files = [file for file in glob.iglob(folder + config["ih"]["file_pattern"], recursive=True)]
    if not files:
        files = [file for file in glob.iglob(folder + '**/*.json', recursive=True)]
        filetype = 'pega_ds_export'

    for file in files:
        if (processed_files["filename"] == file).any():
            logger.info("Skipping file : " + file)
            continue
        try:
            filedate = re.findall(config["ih"]["ih_group_pattern"], os.path.abspath(file))[0]
            if hive_partitioning:
                file_groups[filedate].add(Path(file).parent)
            else:
                file_groups[filedate].add(Path(file))
        except Exception as e:
            file_groups[os.path.basename(file)].add(Path(file))

    file_groups = dict(sorted(file_groups.items(), reverse=True))
    size = len(file_groups.items())

    mdata = {}
    for metric in metrics:
        if db_proxy.is_dataframe_exist(metric):
            mdata[metric] = db_proxy.get_dataframe(metric)

    progress_bar = st.progress(0)
    container = st.empty()
    i = 0
    for key, files_in_grp in file_groups.items():
        logger.debug(f"Processing group: {key}")
        start = time.time()
        i = i + 1
        progress_bar.progress(i / size, text=f"Processing: {key}")
        ih_group = read_file_group(
            list(files_in_grp),
            filetype,
            streaming,
            config,
            hive_partitioning,
            add_columns_expr,
            ih_filter_expr
        )
        if ih_group is None:
            continue
        collect_ih_metrics_data(loop, ih_group, mdata, streaming, background, config, metric_coroutines_map)
        del ih_group

        if (i > 31) & (i % 31 == 1):
            ram_mb = process.memory_info().rss / (1024 * 1024)
            logger.debug(f"RSS = {ram_mb:.2f} MB")
            logger.debug(f"SWAP = {psutil.swap_memory().used / (1024 * 1024):.2f} MB")
        end = time.time()
        logger.debug(f"Time taken: {(end - start) * 10 ** 3:.03f}ms")
        load_mid = time.time()
        hours, remainder = divmod(load_mid - load_start, 3600)
        minutes, seconds = divmod(remainder, 60)
        container.metric("Time", f"{hours:.0f}h:{minutes:.0f}m:{seconds:.02f}s", label_visibility="collapsed")

    progress_bar.empty()
    container.empty()
    collected_metrics_data = {}

    for metric in mdata:
        totals_frame = compact_data(mdata[metric], config['metrics'][metric], metric)
        totals_frame.shrink_to_fit(in_place=True)
        collected_metrics_data[metric] = totals_frame
        db_proxy.drop_dataframe(metric)
        db_proxy.store_dataframe(totals_frame, metric)

    for file in files:
        save_file_meta(db_proxy, file)

    load_end = time.time()
    hours, remainder = divmod(load_end - load_start, 3600)
    minutes, seconds = divmod(remainder, 60)
    logger.info(f"Load time: {hours:.0f}h:{minutes:.0f}m:{seconds:.02f}s")
    for metric in collected_metrics_data:
        frame = collected_metrics_data[metric]
        logger.debug(f"{metric} dataset size is {frame.estimated_size('kb')} kb.")
    st.session_state['drop_cache'] = False
    return collected_metrics_data


def read_file_group(files: typing.List,
                    filetype: str,
                    streaming: bool,
                    config: dict,
                    hive_partitioning: bool,
                    add_columns_expr: typing.Any,
                    ih_filter_expr: typing.Any
                    ) -> LazyFrame | None:
    start: float = time.time()
    if filetype == 'parquet':
        ih = pl.scan_parquet(files, cache=False, hive_partitioning=hive_partitioning, missing_columns='insert',
                             extra_columns='ignore')
    elif filetype == 'pega_ds_export':
        ih = read_dataset_export(files, lazy=True)
    else:
        raise Exception("File type not supported")
    logger.debug(f"Data unpacking and load: {(time.time() - start) * 10 ** 3:.03f}ms")

    dframe_columns = ih.collect_schema().names()
    capitalized = capitalize(dframe_columns)
    rename_map = dict(zip(dframe_columns, capitalized))
    ih = ih.rename(rename_map)

    with_cols_list = []
    if 'default_values' in config["ih"]["extensions"].keys():
        default_values = config["ih"]["extensions"]["default_values"]
        for new_col in default_values.keys():
            if new_col not in capitalized:
                with_cols_list.append(pl.lit(default_values.get(new_col)).alias(new_col))
            else:
                with_cols_list.append(pl.col(new_col).fill_null(default_values.get(new_col)))
    if with_cols_list:
        ih = ih.with_columns(with_cols_list)

    ih = ih.filter(ih_filter_expr)

    ih = (
        ih.with_columns([
            pl.col(OUTCOME_TIME).str.strptime(pl.Datetime, "%Y%m%dT%H%M%S%.3f %Z"),
            pl.col(DECISION_TIME).str.strptime(pl.Datetime, "%Y%m%dT%H%M%S%.3f %Z"),
            pl.concat_str([pl.col(ISSUE), pl.col(GROUP), pl.col(NAME)], separator="/").alias(ACTION_ID),
        ])
        .with_columns([
            pl.col(OUTCOME_TIME).dt.date().alias("Day"),
            pl.col(OUTCOME_TIME).dt.strftime("%Y-%m").alias("Month"),
            pl.col(OUTCOME_TIME).dt.year().cast(pl.Utf8).alias("Year"),
            (pl.col(OUTCOME_TIME).dt.year().cast(pl.Utf8) + "_Q" +
             pl.col(OUTCOME_TIME).dt.quarter().cast(pl.Utf8)).alias("Quarter"),
            (pl.col(OUTCOME_TIME) - pl.col(DECISION_TIME)).dt.total_seconds().alias("ResponseTime")
        ])
        .unique(subset=[INTERACTION_ID, ACTION_ID, RANK, OUTCOME])
        .drop(DROP_IH_COLUMNS, strict=False)
    )
    if add_columns_expr:
        ih = ih.with_columns(add_columns_expr)

    ih = ih.collect(engine="streaming" if streaming else "auto").lazy()
    logger.debug(f"Pre-processing: {(time.time() - start) * 10 ** 3:.03f}ms")
    return ih



import traceback

import polars as pl
from polars import selectors as cs
from polars_ds import weighted_mean

from value_dashboard.utils.config import get_config
from value_dashboard.utils.polars_utils import build_digest, merge_digests
from value_dashboard.utils.string_utils import strtobool
from value_dashboard.utils.timer import timed


@timed
def descriptive(ih: pl.LazyFrame, config: dict, streaming=False, background=False):
    mand_props_grp_by = list(set(config['group_by'] + get_config()["metrics"]["global_filters"]))
    columns = config['columns']
    use_t_digest = strtobool(config['use_t_digest']) if 'use_t_digest' in config.keys() else True
    num_columns = [col for col in columns if col in ih.select(cs.numeric()).collect_schema().names()]

    if "filter" in config:
        filter_exp_cmp = config["filter"]
        if not isinstance(filter_exp_cmp, str):
            ih = ih.filter(config["filter"])

    num_selector = cs.numeric() & cs.by_name(columns, require_all=True)
    common_aggs = [
        pl.col(columns).count().name.suffix('_Count'),
        num_selector.sum().name.suffix('_Sum'),
        num_selector.mean().name.suffix('_Mean'),
        num_selector.var().name.suffix('_Var'),
        num_selector.min().name.suffix('_Min'),
        num_selector.max().name.suffix('_Max')
    ]
    if use_t_digest:
        t_digest_aggs = [
            pl.map_groups(
                exprs=[pl.col(c)],
                function=lambda s: build_digest(s),
                return_dtype=pl.Binary,
                returns_scalar=True
            ).alias(f'{c}_tdigest') for c in num_columns
        ]
        agg_exprs = common_aggs + t_digest_aggs
    else:
        extra_aggs = [
            num_selector.median().name.suffix('_Median'),
            num_selector.skew().name.suffix('_Skew'),
            num_selector.quantile(0.25).name.suffix('_p25'),
            num_selector.quantile(0.75).name.suffix('_p75'),
            num_selector.quantile(0.90).name.suffix('_p90'),
            num_selector.quantile(0.95).name.suffix('_p95')
        ]
        agg_exprs = common_aggs + extra_aggs

    try:
        ih_analysis = ih.group_by(mand_props_grp_by).agg(agg_exprs)
        if background:
            return ih_analysis.collect(background=background, engine="streaming" if streaming else "auto")
        else:
            return ih_analysis
    except Exception as e:
        traceback.print_exc()
        raise e


@timed
def compact_descriptive_data(data: pl.DataFrame,
                             config: dict) -> pl.DataFrame:
    use_t_digest = strtobool(config['use_t_digest']) if 'use_t_digest' in config.keys() else True
    columns_conf = config['columns']
    scores = config['scores']
    num_columns = [col for col in columns_conf if (col + '_Mean') in data.columns]
    grp_by = list(set(config['group_by'] + get_config()["metrics"]["global_filters"]))

    grouped_mean = (
        data
        .group_by(grp_by)
        .agg([pl.col(grp_by).first().name.suffix("_a")]
             +
             [
                 weighted_mean(pl.col(f'{c}_Mean'), pl.col(f'{c}_Count')).alias(f'{c}_GroupMean_a') for c in
                 num_columns
             ]
             )
        .select(cs.ends_with("_a"))
        .rename(lambda column_name: column_name.removesuffix('_a'))
    )

    copy_data = data.join(grouped_mean, on=grp_by)

    copy_data = copy_data.with_columns(
        [((pl.col(f'{c}_Count') - 1) * pl.col(f'{c}_Var')).alias(f'{c}_n_minus1_variance')
         for c in num_columns] +
        [(pl.col(f'{c}_Count') * (pl.col(f'{c}_Mean') - pl.col(f'{c}_GroupMean')) ** 2)
        .alias(f'{c}_n_mean_diff_sq')
         for c in num_columns]
    )

    common_aggs = [
        (cs.ends_with('Count').sum()).name.suffix("_a"),
        (cs.ends_with('Sum').sum()).name.suffix("_a"),
        (cs.ends_with('Min').min()).name.suffix("_a"),
        (cs.ends_with('Max').max()).name.suffix("_a"),
        pl.col(grp_by).first().name.suffix("_a")
    ]

    mean_aggs = [
        weighted_mean(pl.col(f'{c}_Mean'), pl.col(f'{c}_Count')).alias(f'{c}_Mean_a')
        for c in num_columns
    ]

    if not use_t_digest:
        extra_aggs = (
                [weighted_mean(pl.col(f'{c}_Median'), pl.col(f'{c}_Count')).alias(f'{c}_Median_a')
                 for c in num_columns] +
                [weighted_mean(pl.col(f'{c}_Skew'), pl.col(f'{c}_Count')).alias(f'{c}_Skew_a')
                 for c in num_columns]
        )
        if 'p25' in scores:
            extra_aggs += [weighted_mean(pl.col(f'{c}_p25'), pl.col(f'{c}_Count')).alias(f'{c}_p25_a')
                           for c in num_columns]
        if 'p75' in scores:
            extra_aggs += [weighted_mean(pl.col(f'{c}_p75'), pl.col(f'{c}_Count')).alias(f'{c}_p75_a')
                           for c in num_columns]
        if 'p95' in scores:
            extra_aggs += [weighted_mean(pl.col(f'{c}_p95'), pl.col(f'{c}_Count')).alias(f'{c}_p95_a')
                           for c in num_columns]
        if 'p90' in scores:
            extra_aggs += [weighted_mean(pl.col(f'{c}_p90'), pl.col(f'{c}_Count')).alias(f'{c}_p90_a')
                           for c in num_columns]
    else:
        extra_aggs = [
            pl.map_groups(
                exprs=[pl.col(f'{c}_tdigest')],
                function=lambda s: merge_digests(s),
                return_dtype=pl.Binary,
                returns_scalar=True
            ).alias(f'{c}_tdigest_a') for c in num_columns
        ]

    tail_aggs = (
            [pl.col(f'{c}_n_minus1_variance').sum().alias(f'{c}_sum_n_minus1_variance_tmp_a')
             for c in num_columns] +
            [pl.col(f'{c}_n_mean_diff_sq').sum().alias(f'{c}_sum_n_mean_diff_sq_tmp_a')
             for c in num_columns]
    )

    agg_list = common_aggs + mean_aggs + extra_aggs + tail_aggs
    result = (
        copy_data
        .group_by(grp_by)
        .agg(agg_list)
        .select(cs.ends_with("_a"))
        .rename(lambda col: col.removesuffix('_a'))
        .with_columns([
            ((pl.col(f'{c}_sum_n_minus1_variance_tmp') + pl.col(f'{c}_sum_n_mean_diff_sq_tmp'))
             / (pl.col(f'{c}_Count') - 1)).alias(f'{c}_Var')
            for c in num_columns
        ])
        .select(~cs.ends_with("_tmp"))
    )

    return result



import datetime
import traceback

import polars as pl
from dateutil.relativedelta import relativedelta

from value_dashboard.metrics.constants import CLV_MODEL
from value_dashboard.metrics.constants import CUSTOMER_ID
from value_dashboard.metrics.constants import PURCHASED_DATE, ONE_TIME_COST, HOLDING_ID
from value_dashboard.metrics.constants import RECURRING_PERIOD, RECURRING_COST
from value_dashboard.metrics.constants import rfm_config_dict
from value_dashboard.utils.timer import timed


@timed
def clv(holdings: pl.LazyFrame, config: dict, streaming=False, background=False):
    holding_id_col = config['order_id_col'] if 'order_id_col' in config.keys() else HOLDING_ID
    lifespan = config['lifespan'] if 'lifespan' in config.keys() else 2000
    customer_id_col = config['customer_id_col'] if 'customer_id_col' in config.keys() else CUSTOMER_ID
    monetary_value_col = config['monetary_value_col'] if 'monetary_value_col' in config.keys() else ONE_TIME_COST
    purchase_date_col = config['purchase_date_col'] if 'purchase_date_col' in config.keys() else PURCHASED_DATE
    clv_model = config['model'] if 'model' in config.keys() else CLV_MODEL
    recurring_period_col = config['recurring_period'] if 'recurring_period' in config.keys() else RECURRING_PERIOD
    recurring_cost_col = config['recurring_cost'] if 'recurring_cost' in config.keys() else RECURRING_COST
    mand_props_grp_by = config['group_by'] + [customer_id_col, 'Year', 'Quarter']

    if "filter" in config:
        filter_exp_cmp = config["filter"]
        if not isinstance(filter_exp_cmp, str):
            holdings = holdings.filter(config["filter"])

    holdings = holdings.filter(pl.col(purchase_date_col) > (datetime.datetime.now() - relativedelta(years=lifespan)))
    holdings = holdings.with_columns(pl.col(monetary_value_col).cast(pl.Float64))
    try:
        data_aggr = (
            holdings
            .with_columns([
                pl.col(purchase_date_col).dt.date().alias('Day'),
                pl.col(purchase_date_col).dt.strftime('%Y-%m').alias('Month'),
                pl.col(purchase_date_col).dt.year().cast(pl.Utf8).alias('Year'),
                (pl.col(purchase_date_col).dt.year().cast(pl.Utf8) + '_Q' +
                 pl.col(purchase_date_col).dt.quarter().cast(pl.Utf8)).alias('Quarter')
            ])
            .group_by(mand_props_grp_by)
            .agg(
                [
                    pl.col(holding_id_col).n_unique().alias("unique_holdings"),
                    pl.sum(monetary_value_col).alias('lifetime_value'),
                    pl.min(purchase_date_col).alias("MinPurchasedDate"),
                    pl.max(purchase_date_col).alias("MaxPurchasedDate")
                ]
                +
                (
                    [
                        (pl.col(recurring_cost_col) * pl.col(recurring_period_col)).sum().alias("recurring_costs")
                    ] if clv_model == 'contractual' else []
                )
            )
        )
        if clv_model == 'contractual':
            data_aggr = (
                data_aggr
                .with_columns(
                    [
                        (pl.col('recurring_costs') + pl.col('lifetime_value')).alias('lifetime_value_a')
                    ]
                )
                .drop('lifetime_value')
                .rename({'lifetime_value_a': 'lifetime_value'})
            )
        if background:
            return data_aggr.collect(background=background, engine="streaming" if streaming else "auto")
        else:
            return data_aggr
    except Exception as e:
        traceback.print_exc()
        raise e


_default_rfm_segment_config = {
    "Premium Customer": [
        "334",
        "443",
        "444",
        "344",
        "434",
        "433",
        "343",
        "333",
    ],
    "Repeat Customer": ["244", "234", "232", "332", "143", "233", "243", "242"],
    "Top Spender": [
        "424",
        "414",
        "144",
        "314",
        "324",
        "124",
        "224",
        "423",
        "413",
        "133",
        "323",
        "313",
        "134",
    ],
    "At Risk Customer": [
        "422",
        "223",
        "212",
        "122",
        "222",
        "132",
        "322",
        "312",
        "412",
        "123",
        "214",
    ],
    "Inactive Customer": ["411", "111", "113", "114", "112", "211", "311"],
}


def rfm_summary(holdings_aggr: pl.DataFrame, config: dict):
    """Summarize pre-aggregated dataframe for use in CLV calculations and analysis.
    Columns added per customer id:
        frequency, recency, T, monetary_value
    """
    customer_id_col = config['customer_id_col'] if 'customer_id_col' in config.keys() else CUSTOMER_ID
    mand_props_grp_by = config['group_by'] + [customer_id_col]
    rfm_segment_config = rfm_config_dict.get(
        config['rfm_segment_config'] if 'rfm_segment_config' in config.keys() else 'NA', _default_rfm_segment_config)

    observation_period_end_ts = holdings_aggr.select(pl.col("MaxPurchasedDate").max()).item()
    time_scaler = 1.0
    segments_list = [str(x) for x in list(range(1, 5))]
    segments_recency_list = [str(x) for x in list(range(4, 0, -1))]
    segment_names = {}
    for key in rfm_segment_config.keys():
        val_list = rfm_segment_config.get(key)
        for val in val_list:
            segment_names[val] = key

    summary = (
        holdings_aggr
        .group_by(mand_props_grp_by)
        .agg(
            [
                pl.col(customer_id_col).n_unique().alias('customers_count'),
                pl.col("unique_holdings").sum().round(2),
                pl.col('lifetime_value').sum().round(2),
                pl.col("MinPurchasedDate").min(),
                pl.col("MaxPurchasedDate").max()
            ])
        .with_columns(
            [
                (pl.col("unique_holdings") - 1).alias('frequency'),
                ((pl.col("MaxPurchasedDate") - pl.col("MinPurchasedDate")).dt.total_days() / time_scaler).alias(
                    "recency"),
                ((pl.lit(observation_period_end_ts) - pl.col("MinPurchasedDate")).dt.total_days() / time_scaler).alias(
                    'tenure'),
                (pl.col('lifetime_value') / pl.col("unique_holdings")).alias('monetary_value')
            ]
        )
        .with_columns(
            [
                (pl.col('tenure') - pl.col('recency')).alias('recency'),
                pl.when(pl.col('frequency') == 0).then(pl.lit(0.0)).otherwise(pl.col('monetary_value')).alias(
                    'monetary_value')
            ]
        )
        .filter(pl.col(customer_id_col).is_not_null())
        .with_columns(
            pl.col('frequency').qcut(4, labels=segments_list, allow_duplicates=True).alias('f_quartile'),
            pl.col('monetary_value').qcut(4, labels=segments_list, allow_duplicates=True).alias('m_quartile'),
            pl.col('recency').qcut(4, labels=segments_recency_list, allow_duplicates=True).alias('r_quartile')
        )
        .with_columns(
            pl.concat_str(
                [
                    pl.col("r_quartile"),
                    pl.col("f_quartile"),
                    pl.col("m_quartile")
                ],
                separator="",
            ).alias("rfm_seg")
        )
        .with_columns(
            pl.col("rfm_seg").replace(segment_names, default="Unknown").alias("rfm_segment")
        )
        .with_columns(
            pl.mean_horizontal(
                [
                    pl.col("r_quartile").cast(pl.String).str.to_decimal(),
                    pl.col("f_quartile").cast(pl.String).str.to_decimal(),
                    pl.col("m_quartile").cast(pl.String).str.to_decimal()
                ]
            ).round(2).alias("rfm_score")
        )
        .sort(mand_props_grp_by, descending=True)
        .drop(["MinPurchasedDate", "MaxPurchasedDate", "rfm_seg"])
    )
    return summary



import asyncio
import typing
from collections import OrderedDict
from typing import Any

import polars as pl
from polars import DataFrame

from value_dashboard.metrics.clv import clv
from value_dashboard.metrics.conversion import compact_conversion_data
from value_dashboard.metrics.descriptive import compact_descriptive_data
from value_dashboard.metrics.engagement import compact_engagement_data
from value_dashboard.metrics.experiment import compact_experiment_data
from value_dashboard.metrics.ml import compact_model_ml_scores_data
from value_dashboard.utils.config import get_config
from value_dashboard.utils.logger import get_logger
from value_dashboard.utils.timer import timed

logger = get_logger(__name__)
data_cache_hours = 24
if 'data_cache_hours' in get_config()['ux'].keys():
    data_cache_hours = get_config()['ux']['data_cache_hours']
logger.debug(f"Data will be cached for {data_cache_hours} hours.")


async def run_data_collection_async(*coros_or_futures):
    result = await asyncio.gather(*coros_or_futures)
    return result


@timed
def collect_ih_metrics_data(loop, ih: pl.DataFrame | pl.LazyFrame,
                            mdata: typing.Dict[
                                str,
                                pl.LazyFrame |
                                pl.DataFrame |
                                pl.lazyframe.in_process.InProcessQuery
                            ],
                            streaming: bool,
                            background: bool,
                            config: dict,
                            metric_coroutines_map: dict):
    metrics = config["metrics"]
    global_filters = metrics["global_filters"]
    coroutines = []
    for metric in metric_coroutines_map.keys():
        params = metrics[metric]
        add_grp_by = list(set(global_filters) - set(params['group_by']))
        params['group_by'] = params['group_by'] + add_grp_by
        coroutines.append(
            data_collection_async(
                ih, params, metric, streaming, background, metric_coroutines_map.get(metric)
            )
        )
    process_metrics_coroutines(coroutines, loop, mdata, streaming)


@timed
def collect_clv_metrics_data(loop, holdings: pl.DataFrame | pl.LazyFrame,
                             mdata: typing.Dict[
                                 str,
                                 pl.LazyFrame |
                                 pl.DataFrame |
                                 pl.lazyframe.in_process.InProcessQuery
                             ],
                             streaming: bool,
                             background: bool,
                             config: dict):
    metrics = config["metrics"]
    coroutines = []
    for metric in metrics:
        params = metrics[metric]
        if metric.startswith("clv"):
            coroutines.append(data_collection_async(holdings, params, metric, streaming, background, clv))
    process_metrics_coroutines(coroutines, loop, mdata, streaming)


def process_metrics_coroutines(coroutines: typing.List, loop,
                               mdata: typing.Dict[
                                   str,
                                   pl.LazyFrame |
                                   pl.DataFrame |
                                   pl.lazyframe.in_process.InProcessQuery
                               ],
                               streaming: bool):
    result = loop.run_until_complete(run_data_collection_async(*coroutines))
    lazy_frames = OrderedDict()
    for metric, mdf in result:
        if isinstance(mdf, pl.lazyframe.in_process.InProcessQuery):
            df = mdf.fetch_blocking()
            schema = df.collect_schema()
            df = pl.from_arrow(df.to_arrow(), schema=schema)
            if not (metric in mdata):
                mdata[metric] = df
            else:
                mdata[metric] = pl.concat([df, mdata[metric]], how="diagonal", rechunk=True)
        elif isinstance(mdf, pl.LazyFrame):
            lazy_frames[metric] = mdf
        else:
            if not (metric in mdata):
                mdata[metric] = mdf
            else:
                schema = mdf.collect_schema()
                mdf = pl.from_arrow(mdf.to_arrow(), schema=schema)
                mdata[metric] = pl.concat([mdf, mdata[metric]], how="diagonal", rechunk=True)
    if lazy_frames:
        frames = pl.collect_all(lazy_frames.values(), engine="streaming" if streaming else "auto")
        for metric in lazy_frames.keys():
            df = frames[list(lazy_frames.keys()).index(metric)]
            schema = df.collect_schema()
            df = pl.from_arrow(df.to_arrow(), schema=schema)
            if not (metric in mdata):
                mdata[metric] = df
            else:
                mdata[metric] = pl.concat([df, mdata[metric]], how="diagonal", rechunk=True)


@timed
def collect_reports_data(collected_metrics_data: typing.Dict[str, pl.DataFrame]) -> dict[Any, tuple[DataFrame, Any]]:
    report_params = get_config()["reports"]
    reports_data: dict[Any, tuple[DataFrame, Any]] = {}
    for report in report_params:
        params = report_params[report]
        if params['metric'].startswith("engagement"):
            report_data = collected_metrics_data[params['metric']].clone()
            reports_data[report] = (report_data, params)
        elif params['metric'].startswith("model_ml_scores"):
            report_data = collected_metrics_data[params['metric']].clone()
            reports_data[report] = (report_data, params)
        elif params['metric'].startswith("conversion"):
            report_data = collected_metrics_data[params['metric']].clone()
            reports_data[report] = (report_data, params)
        elif params['metric'].startswith("descriptive"):
            report_data = collected_metrics_data[params['metric']].clone()
            reports_data[report] = (report_data, params)
        elif params['metric'].startswith("experiment"):
            report_data = collected_metrics_data[params['metric']].clone()
            reports_data[report] = (report_data, params)
    return reports_data


@timed
def get_reports_data_by_name(report_name: str, collected_metrics_data: typing.Dict[str, pl.DataFrame]) -> tuple[
    pl.DataFrame, dict]:
    report_params = get_config()["reports"]
    reports_data: tuple[DataFrame, Any] = ()
    params = report_params[report_name]
    if params['metric'].startswith("engagement"):
        report_data = collected_metrics_data[params['metric']].clone()
        reports_data = (report_data, params)
    elif params['metric'].startswith("model_ml_scores"):
        report_data = collected_metrics_data[params['metric']].clone()
        reports_data = (report_data, params)
    elif params['metric'].startswith("conversion"):
        report_data = collected_metrics_data[params['metric']].clone()
        reports_data = (report_data, params)
    elif params['metric'].startswith("descriptive"):
        report_data = collected_metrics_data[params['metric']].clone()
        reports_data = (report_data, params)
    elif params['metric'].startswith("experiment"):
        report_data = collected_metrics_data[params['metric']].clone()
        reports_data = (report_data, params)
    return reports_data


async def data_collection_async(ih: pl.DataFrame,
                                config: dict,
                                metric: str,
                                streaming: bool,
                                background: bool,
                                func: typing.Callable):
    return metric, func(ih, config, streaming, background)


@timed
def compact_data(data: pl.DataFrame, params: dict, metric: str) -> pl.DataFrame:
    if metric.startswith("engagement"):
        data = compact_engagement_data(data, params)
    elif metric.startswith("model_ml_scores"):
        data = compact_model_ml_scores_data(data, params)
    elif metric.startswith("conversion"):
        data = compact_conversion_data(data, params)
    elif metric.startswith("descriptive"):
        data = compact_descriptive_data(data, params)
    elif metric.startswith("experiment"):
        data = compact_experiment_data(data, params)
    return data



import numpy as np
import plotly.graph_objs as go
import polars_ds.sample_and_split as pds
from lifetimes import BetaGeoFitter, ParetoNBDFitter, GammaGammaFitter
from plotly.subplots import make_subplots

from value_dashboard.metrics.constants import CUSTOMER_ID
from value_dashboard.reports.shared_plot_utils import *
from value_dashboard.utils.config import get_config


@timed
def clv_histogram_plot(data: Union[pl.DataFrame, pd.DataFrame],
                       config: dict) -> pd.DataFrame:
    report_data = calculate_reports_data(data, config).to_pandas()
    rep_filtered_data = filter_dataframe(align_column_types(report_data), case=False)
    if rep_filtered_data.shape[0] == 0:
        st.warning("No data available.")
        st.stop()

    c1, c2, c3, c4, c5 = st.columns(5)
    with c1:
        options_x = ['frequency', 'recency', 'monetary_value', 'tenure']
        config['x'] = st.selectbox(
            label='X-Axis',
            options=options_x,
            index=options_x.index(config['x']),
            help="Select X-Axis value."
        )
    with c2:
        options_histnorm = ['', 'percent', 'probability', 'density', 'probability density']
        histnorm = st.selectbox(
            label='Normalization',
            options=options_histnorm,
            index=options_histnorm.index(''),
            help="Select histnorm option for the plot."
        )
    with c3:
        options_y = [None, 'lifetime_value', 'unique_holdings']
        y = st.selectbox(
            label='Y-Axis',
            options=options_y,
            index=options_y.index(None),
            help="Select Y-Axis value."
        )
    with c4:
        options_histfunc = ['count', 'sum', 'avg', 'min', 'max']
        histfunc = st.selectbox(
            label='Y-Axis Aggregation',
            options=options_histfunc,
            index=options_histfunc.index('count'),
            help="Select histfunc option for the plot."
        )
    with c5:
        cumulative = st.radio(
            'Cumulative',
            (False, True),
            horizontal=True
        )

    if 'facet_row' in config.keys():
        height = max(640, 300 * len(rep_filtered_data[config['facet_row']].unique()))
    else:
        height = 640

    fig = px.histogram(
        rep_filtered_data,
        x=config['x'],
        y=y,
        histnorm=histnorm,
        histfunc=histfunc,
        cumulative=cumulative,
        facet_col=config['facet_column'] if 'facet_column' in config.keys() else None,
        facet_row=config['facet_row'] if 'facet_row' in config.keys() else None,
        color=config['color'] if 'color' in config.keys() else None,
        title=config['description'],
        height=height,
        text_auto=True,
        marginal="box",
        # barmode='group'
    )
    fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
    fig.update_traces(marker_line_width=1, marker_line_color="white")
    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    return rep_filtered_data


@timed
def clv_polarbar_plot(data: Union[pl.DataFrame, pd.DataFrame],
                      config: dict) -> pd.DataFrame:
    clv_totals_cards_subplot(data, config)
    data = calculate_reports_data(data, config).to_pandas()
    c1, c2, c3 = st.columns(3)
    with c1:
        options_r = ['lifetime_value', 'unique_holdings', 'monetary_value']
        config['r'] = st.selectbox(
            label='Radial-Axis',
            options=options_r,
            index=options_r.index(config['r']),
            help="Select Radial-Axis value."
        )
    with c2:
        options_theta = list(set(['rfm_segment'] + config['group_by']))
        config['theta'] = st.selectbox(
            label='Angular axis in polar coordinates',
            options=options_theta,
            index=options_theta.index(config['theta']),
            help="Select  angular axis in polar coordinates."
        )
    with c3:
        options_color = list(set(['rfm_segment', 'f_quartile', 'r_quartile', 'm_quartile'] + config['group_by']))
        config['color'] = st.selectbox(
            label='Colour',
            options=options_color,
            index=options_color.index(config['color']),
            help="Select colour value."
        )

    grp_by = []
    if not config['theta'] in grp_by:
        grp_by.append(config['theta'])
    if not config['color'] in grp_by:
        grp_by.append(config['color'])

    if grp_by:
        if isinstance(data, pd.DataFrame):
            data = pl.from_pandas(data)
        report_data = (
            data
            .group_by(grp_by)
            .agg(
                pl.sum("customers_count").alias("Customers count"),
                pl.sum("lifetime_value").alias("lifetime_value"),
                pl.sum("unique_holdings").alias("unique_holdings"),
                pl.mean("monetary_value").alias("monetary_value"),
                pl.mean("frequency").alias("Avg frequency"),
                pl.mean("rfm_score").alias("Avg rfm score")
            )
            .sort(grp_by)
        )
    else:
        report_data = data
    report_data = report_data.to_pandas()
    ih_analysis = filter_dataframe(align_column_types(report_data), case=False)

    if ih_analysis.shape[0] == 0:
        st.warning("No data available.")
        st.stop()
    theme = st.context.theme.type
    if theme is None:
        template = 'none'
    else:
        if theme == 'dark':
            template = 'plotly_dark'
        else:
            template = 'none'
    fig = px.bar_polar(ih_analysis,
                       r=config["r"],
                       theta=config["theta"],
                       color=config["color"],
                       barmode="group",
                       template=template,
                       title=config['description']
                       )
    fig.update_layout(
        polar_hole=0.25,
        height=700,
        # width=1400,
        margin=dict(b=25, t=50, l=0, r=0),
        showlegend=strtobool(config["showlegend"])
    )
    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    return ih_analysis


@timed
def clv_treemap_plot(data: Union[pl.DataFrame, pd.DataFrame],
                     config: dict) -> pd.DataFrame:
    clv_totals_cards_subplot(data, config)
    data = calculate_reports_data(data, config)
    grp_by = ['rfm_segment'] + config['group_by']
    if grp_by:
        if isinstance(data, pd.DataFrame):
            data = pl.from_pandas(data)
        report_data = (
            data
            .group_by(grp_by)
            .agg(
                pl.sum("customers_count").round(2).alias("Customers count"),
                pl.sum("lifetime_value").round(2).alias("lifetime_value"),
                pl.sum("unique_holdings").round(2).alias("unique_holdings"),
                pl.mean("monetary_value").round(2).alias("monetary_value"),
                pl.mean("frequency").round(2).alias("Avg frequency"),
                pl.mean("rfm_score").round(2).alias("Avg rfm score")
            )
            .sort(grp_by)
        )
    else:
        report_data = data
    report_data = report_data.to_pandas()
    ih_analysis = filter_dataframe(align_column_types(report_data), case=False)

    if ih_analysis.shape[0] == 0:
        st.warning("No data available.")
        st.stop()
    fig = px.treemap(ih_analysis, path=[px.Constant(" ")] + grp_by, values='Customers count',
                     color="Avg rfm score",
                     color_continuous_scale=["#D61F1F", "#E03C32", "#FFD301", "#639754", "#006B3D"],
                     title=config['description'],
                     hover_data=['lifetime_value', 'unique_holdings', 'monetary_value', "Avg frequency",
                                 "Avg rfm score"],
                     height=640,
                     )
    fig.update_traces(textinfo="label+text+value+percent root", root_color="lightgrey")
    fig.update_layout(margin=dict(t=50, l=25, r=25, b=25))
    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    return ih_analysis


@timed
def clv_exposure_plot(data: Union[pl.DataFrame, pd.DataFrame],
                      config: dict) -> pd.DataFrame:
    data = calculate_reports_data(data, config).to_pandas()
    clv_analysis = pl.from_pandas(filter_dataframe(align_column_types(data), case=False))

    if clv_analysis.shape[0] == 0:
        st.warning("No data available.")
        st.stop()

    clv_analysis = pds.sample(clv_analysis, 100).sort(['recency', 'tenure'])
    fig = clv_plot_customer_exposure(clv_analysis, linewidth=0.5, size=0.75)
    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    return clv_analysis.to_pandas()


@timed
def clv_correlation_plot(data: Union[pl.DataFrame, pd.DataFrame],
                         config: dict) -> pd.DataFrame:
    data = calculate_reports_data(data, config).to_pandas()
    c1, c2, c3, c4 = st.columns(4)
    with c1:
        options_par1 = ['recency', 'frequency', 'monetary_value', 'tenure', 'lifetime_value']
        config['x'] = st.selectbox(
            label='X-Axis',
            options=options_par1,
            index=options_par1.index(config['x']),
            help="Select X-Axis value."
        )
    with c2:
        options_par2 = ['recency', 'frequency', 'monetary_value', 'tenure', 'lifetime_value']
        config['y'] = st.selectbox(
            label='Y-Axis',
            options=options_par2,
            index=options_par2.index(config['y']),
            help="Select Y-Axis value."
        )
    with c3:
        options_facet_col = [None, 'rfm_segment', 'ControlGroup']
        if 'facet_col' in config.keys():
            config['facet_col'] = st.selectbox(
                label='Group By',
                options=options_facet_col,
                index=options_facet_col.index(config['facet_col']),
                help="Select Group By value."
            )
        else:
            config['facet_col'] = st.selectbox(
                label='Group By',
                options=options_facet_col,
                help="Select Group By value."
            )
    with c4:
        method = st.selectbox(
            label='Correlation method',
            options=['pearson', 'kendall', 'spearman'],
            help="""Method used to compute correlation:
- pearson : Standard correlation coefficient
- kendall : Kendall Tau correlation coefficient
- spearman : Spearman rank correlation"""
        )
    ih_analysis = filter_dataframe(align_column_types(data), case=False)

    if ih_analysis.shape[0] == 0:
        st.warning("No data available.")
        st.stop()

    if config['facet_col']:
        facets = sorted(ih_analysis[config['facet_col']].unique())
        img_sequence = []
        for facet_col in facets:
            img_sequence.append(
                ih_analysis[ih_analysis[config['facet_col']] == facet_col][[config['x'], config['y']]]
                .corr(method=method)
            )
        img_sequence = np.array(img_sequence)

    else:
        img_sequence = ih_analysis[[config['x'], config['y']]].corr(method=method)

    fig = px.imshow(
        img_sequence,
        color_continuous_scale='Viridis',
        text_auto=".4f",
        aspect='auto',
        x=[config['x'], config['y']],
        y=[config['x'], config['y']],
        facet_col=0 if config['facet_col'] else None,
        facet_col_wrap=4 if config['facet_col'] else None,
    )
    if config['facet_col']:
        for i, label in enumerate(facets):
            fig.layout.annotations[i]['text'] = label
    fig.update_layout(
        title=method.title() + " correlation between " + config['x'] + " and " + config['y']
    )
    st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    return ih_analysis


@timed
def clv_totals_cards_subplot(clv_analysis: Union[pl.DataFrame, pd.DataFrame],
                             config: dict):
    if isinstance(clv_analysis, pd.DataFrame):
        clv_analysis = pl.from_pandas(clv_analysis)

    total_data = calculate_reports_data(clv_analysis, config)

    m_config = get_config()["metrics"][config["metric"]]
    customer_id_col = (
        m_config["customer_id_col"]
        if "customer_id_col" in m_config.keys()
        else CUSTOMER_ID
    )

    num_cols = 4
    cols = st.columns(num_cols, vertical_alignment='center')
    unique_customers = total_data.select(pl.col(customer_id_col).n_unique())
    total_value = total_data.select(pl.col("lifetime_value").sum())
    cols[0].metric(label='Unique customers', value='{:,}'.format(unique_customers.item()).replace(",", " "))
    cols[1].metric(label='Total value', value='{:,.2f}'.format(total_value.item()))

    years = clv_analysis.select("Year").unique().sort("Year")["Year"].to_list()
    if len(years) < 3:
        return
    year1, year2, cur_year = years[-3], years[-2], years[-1],

    df_last_two = clv_analysis.filter(pl.col("Year").is_in([year1, year2]))
    avg_per_year = (df_last_two.group_by("Year")
                    .agg((pl.col("lifetime_value").sum() / pl.col(customer_id_col).n_unique()).alias("avg"))
                    )
    avg_sorted = avg_per_year.sort("Year")
    avg1, avg2 = avg_sorted["avg"].to_list()
    percentage_diff = ((avg2 - avg1) / avg1)

    cols[2].metric(label=year2 + ' average CLTV', value='{:,.2f}'.format(avg2),
                   delta='{:.2%} YoY'.format(percentage_diff))

    cur_df = clv_analysis.filter(pl.col("Year").is_in([cur_year]))
    avg_per_year = (cur_df.group_by("Year")
                    .agg((pl.col("lifetime_value").sum() / pl.col(customer_id_col).n_unique()).alias("avg"))
                    )
    avg_sorted = avg_per_year.sort("Year")
    avg, = avg_sorted["avg"].to_list()
    percentage_diff = ((avg - avg2) / avg2)
    cols[3].metric(label=cur_year + ' average CLTV', value='{:,.2f}'.format(avg),
                   delta='{:.2%} YoY'.format(percentage_diff))


@timed
def clv_model_plot(data: Union[pl.DataFrame, pd.DataFrame],
                   config: dict) -> pd.DataFrame:
    clv_totals_cards_subplot(data, config)
    clv = calculate_reports_data(data, config).to_pandas()
    clv = clv[clv['frequency'] > 0]
    c1, c2 = st.columns(2)
    with c1:
        options_model = ['Gamma - Gamma Model', 'BG/NBD Model', 'Pareto/NBD model']
        model = st.selectbox(
            label='LTV prediction model',
            options=options_model,
            help="Select LTV prediction model."
        )

    with c2:
        lifespan = [1, 2, 3, 5, 8]
        predict_lifespan = st.selectbox(
            label='Predict LTV in years',
            options=lifespan,
            help="Select LTV prediction time."
        )
    bgf = BetaGeoFitter(penalizer_coef=0.001)
    bgf.fit(clv['frequency'], clv['recency'], clv['tenure'], verbose=True)
    t = 12 * 30 * predict_lifespan
    clv['expected_number_of_purchases'] = bgf.conditional_expected_number_of_purchases_up_to_time(t, clv['frequency'],
                                                                                                  clv['recency'],
                                                                                                  clv['tenure'])
    if model == 'BG/NBD Model':
        clv_plt = clv.groupby('rfm_segment')['expected_number_of_purchases'].mean().reset_index()
        fig = px.bar(clv_plt,
                     x='rfm_segment',
                     y='expected_number_of_purchases',
                     color='rfm_segment',
                     )

        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
        st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    elif model == 'Pareto/NBD model':
        with st.spinner("Wait for it...", show_time=True):
            pnbmf = ParetoNBDFitter(penalizer_coef=0.001)
            pnbmf.fit(clv['frequency'], clv['recency'], clv['tenure'], verbose=True, maxiter=200)
            clv['expected_number_of_purchases'] = pnbmf.conditional_expected_number_of_purchases_up_to_time(t,
                                                                                                            clv[
                                                                                                                'frequency'],
                                                                                                            clv[
                                                                                                                'recency'],
                                                                                                            clv[
                                                                                                                'tenure'])
            clv_plt = clv.groupby('rfm_segment')['expected_number_of_purchases'].mean().reset_index()
            fig = px.bar(clv_plt,
                         x='rfm_segment',
                         y='expected_number_of_purchases',
                         color='rfm_segment',
                         )

            fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
            st.plotly_chart(fig, use_container_width=True, theme="streamlit")
    else:
        with st.spinner("Wait for it...", show_time=True):
            ggf = GammaGammaFitter(penalizer_coef=0.01)
            ggf.fit(clv["frequency"], clv["monetary_value"], verbose=True)
            clv["expected_lifetime_value"] = ggf.customer_lifetime_value(
                bgf,
                clv["frequency"],
                clv["recency"],
                clv["tenure"],
                clv["monetary_value"],
                time=12 * predict_lifespan,
                freq="D",
                discount_rate=0.01,
            )
            clv_plt = clv.groupby('rfm_segment')['expected_lifetime_value'].mean().reset_index()
            fig = px.bar(clv_plt,
                         x='rfm_segment',
                         y='expected_lifetime_value',
                         color='rfm_segment',
                         )

            fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
            st.plotly_chart(fig, use_container_width=True, theme="streamlit")

    return clv


def clv_plot_customer_exposure(
        df: pl.DataFrame,
        linewidth: float | None = None,
        size: float | None = None,
        labels: list[str] | None = None,
        colors: list[str] | None = None,
        padding: float = 0.25
) -> go.Figure:
    if padding < 0:
        raise ValueError("padding must be non-negative")

    if size is not None and size < 0:
        raise ValueError("size must be non-negative")

    if linewidth is not None and linewidth < 0:
        raise ValueError("linewidth must be non-negative")

    n = len(df)
    customer_idx = list(range(1, n + 1))

    recency = df['recency'].to_list()
    T = df['tenure'].to_list()

    if colors is None:
        colors = ["blue", "orange"]

    if len(colors) != 2:
        raise ValueError("colors must be a sequence of length 2")

    recency_color, T_color = colors
    fig = make_subplots()
    for idx, (rec, t) in enumerate(zip(recency, T)):
        fig.add_trace(
            go.Scatter(
                x=[0, rec],
                y=[customer_idx[idx], customer_idx[idx]],
                mode='lines',
                line=dict(color=recency_color, width=linewidth)
            )
        )

    for idx, (rec, t) in enumerate(zip(recency, T)):
        fig.add_trace(
            go.Scatter(
                x=[rec, t],
                y=[customer_idx[idx], customer_idx[idx]],
                mode='lines',
                line=dict(color=T_color, width=linewidth)
            )
        )
    fig.add_trace(
        go.Scatter(
            x=recency,
            y=customer_idx,
            mode='markers',
            marker=dict(color=recency_color, size=size),
            name=labels[0] if labels else 'Recency'
        )
    )
    fig.add_trace(
        go.Scatter(
            x=T,
            y=customer_idx,
            mode='markers',
            marker=dict(color=T_color, size=size),
            name=labels[1] if labels else 'tenure'
        )
    )

    fig.update_layout(
        title="Customer Exposure",
        xaxis_title="Time since first purchase",
        yaxis_title="Customer",
        xaxis=dict(range=[-padding, max(T) + padding]),
        yaxis=dict(range=[1 - padding, n + padding]),
        showlegend=False,
        barmode='group',
        height=640
    )

    return fig
